"""Terraform version management."""

from __future__ import annotations

import hashlib
import json
import locale
import logging
import os
import platform
import re
import shutil
import subprocess
import sys
import tempfile
import zipfile
from typing import TYPE_CHECKING, Any, Final, cast, overload
from urllib.error import URLError
from urllib.request import urlretrieve

import hcl
import hcl2
import requests
from packaging.version import InvalidVersion

from ..compat import cached_property
from ..exceptions import HclParserError
from ..utils import FileHash, Version, get_hash_for_filename, merge_dicts
from . import EnvManager, handle_bin_download_error

if TYPE_CHECKING:
    from collections.abc import Generator
    from pathlib import Path
    from types import ModuleType

    from .._logging import RunwayLogger

LOGGER = cast("RunwayLogger", logging.getLogger(__name__))
TF_VERSION_FILENAME = ".terraform-version"


def download_tf_release(
    version: str,
    versions_dir: Path,
    command_suffix: str,
    tf_platform: str | None = None,
    arch: str | None = None,
) -> None:
    """Download Terraform archive and return path to it."""
    version_dir = versions_dir / version

    if arch is None:
        arch = os.getenv("TFENV_ARCH", "amd64")

    if tf_platform:
        tfver_os = tf_platform + "_" + arch
    elif platform.system().startswith("Darwin"):
        tfver_os = f"darwin_{arch}"
    elif platform.system().startswith("Windows") or (
        platform.system().startswith("MINGW64")
        or (platform.system().startswith("MSYS_NT") or platform.system().startswith("CYGWIN_NT"))
    ):
        tfver_os = f"windows_{arch}"
    else:
        tfver_os = f"linux_{arch}"

    download_dir = tempfile.mkdtemp()
    filename = f"terraform_{version}_{tfver_os}.zip"
    shasums_name = f"terraform_{version}_SHA256SUMS"
    tf_url = "https://releases.hashicorp.com/terraform/" + version

    try:
        LOGGER.verbose("downloading Terraform from %s...", tf_url)
        for i in [filename, shasums_name]:
            urlretrieve(tf_url + "/" + i, os.path.join(download_dir, i))  # noqa: PTH118, S310
    except URLError as exc:
        handle_bin_download_error(exc, "Terraform")

    tf_hash = get_hash_for_filename(
        filename, os.path.join(download_dir, shasums_name)  # noqa: PTH118
    )
    checksum = FileHash(hashlib.sha256())
    checksum.add_file(os.path.join(download_dir, filename))  # noqa: PTH118
    if tf_hash != checksum.hexdigest:
        LOGGER.error("downloaded Terraform %s does not match sha256 %s", filename, tf_hash)
        sys.exit(1)

    with zipfile.ZipFile(os.path.join(download_dir, filename)) as tf_zipfile:  # noqa: PTH118
        version_dir.mkdir(parents=True, exist_ok=True)
        tf_zipfile.extractall(str(version_dir))

    shutil.rmtree(download_dir)
    result = version_dir / ("terraform" + command_suffix)
    result.chmod(result.stat().st_mode | 0o0111)  # ensure it is executable


def get_available_tf_versions(include_prerelease: bool = False) -> list[str]:
    """Return available Terraform versions."""
    tf_releases = json.loads(
        requests.get("https://releases.hashicorp.com/index.json", timeout=30).text
    )["terraform"]

    # Remove versions that don't align with
    # PEP440 (https://peps.python.org/pep-0440/)
    for k, _v in list(tf_releases["versions"].items()):
        try:
            Version(k)
        except InvalidVersion:
            LOGGER.debug("InvalidVersion found, skipping %s. ", k)
            del tf_releases["versions"][k]

    tf_versions = sorted(
        [k for k, _v in tf_releases["versions"].items()],  # descending
        key=Version,
        reverse=True,
    )

    if include_prerelease:
        return [i for i in tf_versions if i]
    return [i for i in tf_versions if i and "-" not in i]


def get_latest_tf_version(include_prerelease: bool = False) -> str:
    """Return latest Terraform version."""
    return get_available_tf_versions(include_prerelease)[0]


def load_terraform_module(parser: ModuleType, path: Path) -> dict[str, Any]:
    """Load all Terraform files in a module into one dict.

    Args:
        parser: Parser to use when loading files.
        path: Terraform module path. All Terraform files in the
            path will be loaded.

    """
    result: dict[str, Any] = {}
    LOGGER.debug("using %s parser to load module: %s", parser.__name__.upper(), path)
    for tf_file in path.glob("*.tf"):
        try:
            tf_config = parser.loads(tf_file.read_text())
            result = merge_dicts(result, cast("dict[str, Any]", tf_config))
        except Exception as exc:  # noqa: BLE001
            raise HclParserError(exc, tf_file, parser) from None
    return result


class TFEnvManager(EnvManager):
    """Terraform version management.

    Designed to be compatible with https://github.com/tfutils/tfenv.

    """

    VERSION_REGEX: Final[str] = r"^(Terraform v)?(?P<version>[0-9]+\.[0-9]+\.[0-9]+\S*)"
    VERSION_OUTPUT_REGEX: Final[str] = (
        r"^Terraform v(?P<version>[0-9]*\.[0-9]*\.[0-9]*)(?P<suffix>-.*)?"
    )

    def __init__(self, path: Path | None = None) -> None:
        """Initialize class."""
        super().__init__("terraform", "tfenv", path)

    @cached_property
    def backend(self) -> dict[str, Any]:
        """Backend config of the Terraform module."""
        # Terraform can only have one backend configured; this formats the
        # data to make it easier to work with
        return next(
            {"type": k, "config": v}
            for k, v in self.terraform_block.get(
                "backend", {None: cast("dict[str, str]", {})}
            ).items()
        )

    @cached_property
    def terraform_block(self) -> dict[str, Any]:  # noqa: C901
        """Collect Terraform configuration blocks from a Terraform module."""

        @overload
        def _flatten_lists(data: dict[str, Any]) -> dict[str, Any]: ...

        @overload
        def _flatten_lists(data: list[Any]) -> list[Any]: ...

        @overload
        def _flatten_lists(data: str) -> str: ...

        def _flatten_lists(data: dict[str, Any] | list[Any] | Any) -> dict[str, Any] | Any:
            """Flatten HCL2 list attributes until its fixed.

            python-hcl2 incorrectly turns all attributes into lists so we need
            to flatten them so they are more similar to HCL.

            https://github.com/amplify-education/python-hcl2/issues/6

            Args:
                data: Dict with lists to flatten.

            """
            if not isinstance(data, dict):
                return data
            copy_data = data.copy()
            for attr, val in copy_data.items():
                if isinstance(val, list):
                    if len(cast("list[Any]", val)) == 1:
                        # pull single values out of lists
                        data[attr] = _flatten_lists(cast(Any, val[0]))
                    else:
                        data[attr] = [_flatten_lists(v) for v in cast("list[Any]", val)]
                elif isinstance(val, dict):
                    data[attr] = _flatten_lists(cast("dict[str, Any]", val))
            return data

        try:
            result: dict[str, Any] | list[dict[str, Any]] = load_terraform_module(
                hcl2, self.path
            ).get("terraform", cast("dict[str, Any]", {}))
        except HclParserError as exc:
            LOGGER.warning(exc)
            LOGGER.warning("failed to parse as HCL2; trying HCL...")
            try:
                result = load_terraform_module(hcl, self.path).get(
                    "terraform", cast("dict[str, Any]", {})
                )
            except HclParserError as exc2:
                LOGGER.warning(exc2)
                # return an empty dict if we can't parse HCL
                # let Terraform decide if it's actually valid
                result = {}

        # python-hcl2 turns all blocks into lists in v0.3.0. this flattens it.
        if isinstance(result, list):
            return _flatten_lists({k: v for i in result for k, v in i.items()})
        return _flatten_lists(result)

    @cached_property
    def version(self) -> Version | None:
        """Terraform version."""
        version_requested = self.current_version or self.get_version_from_file()

        if not version_requested:
            return None

        if re.match(r"^min-required$", version_requested):
            LOGGER.debug("tfenv: detecting minimal required version")
            version_requested = self.get_min_required()

        if re.match(r"^latest:.*$", version_requested):
            regex = re.search(r"latest:(.*)", version_requested).group(1)  # type: ignore
            include_prerelease_versions = False
        elif re.match(r"^latest$", version_requested):
            regex = r"^[0-9]+\.[0-9]+\.[0-9]+$"
            include_prerelease_versions = False
        else:
            regex = f"^{version_requested}$"
            include_prerelease_versions = True
            # Return early (i.e before reaching out to the internet) if the
            # matching version is already installed
            if (self.versions_dir / version_requested).is_dir():
                self.current_version = version_requested
                return self.parse_version_string(self.current_version)

        try:
            version = next(
                i
                for i in get_available_tf_versions(include_prerelease_versions)
                if re.match(regex, i)
            )
        except StopIteration:
            LOGGER.error("unable to find a Terraform version matching regex: %s", regex)
            sys.exit(1)
        self.current_version = version
        return self.parse_version_string(self.current_version)

    @cached_property
    def version_file(self) -> Path | None:
        """Find and return a ".terraform-version" file if one is present.

        Returns:
            Path to the Terraform version file.

        """
        for path in [self.path, self.path.parent]:
            test_path = path / TF_VERSION_FILENAME
            if test_path.is_file():
                LOGGER.debug("using version file: %s", test_path)
                return test_path
        return None

    def get_min_required(self) -> str:
        """Get the defined minimum required version of Terraform.

        Returns:
            The minimum required version as defined in the module.

        """
        version = self.terraform_block.get("required_version")

        if version:
            if re.match(r"^!=.+", version):
                LOGGER.error(
                    "min required Terraform version is a negation (%s) "
                    "- unable to determine required version",
                    version,
                )
                sys.exit(1)
            else:
                version = re.search(r"[0-9]*\.[0-9]*(?:\.[0-9]*)?", version)
                if version:
                    LOGGER.debug("detected minimum Terraform version is %s", version)
                    return version.group(0)
        LOGGER.error(
            "Terraform version specified as min-required, but unable to "
            "find a specified version requirement in this module's tf files"
        )
        sys.exit(1)

    def get_version_from_file(self, file_path: Path | None = None) -> str | None:
        """Get Terraform version from a file.

        Args:
            file_path: Path to file that will be read.

        """
        file_path = file_path or self.version_file
        if file_path and file_path.is_file():
            return file_path.read_text(
                encoding=locale.getpreferredencoding(do_setlocale=False)
            ).strip()
        LOGGER.debug("file path not provided and version file could not be found")
        return None

    def install(self, version_requested: str | None = None) -> str:
        """Ensure Terraform is available."""
        if version_requested:
            self.set_version(version_requested)

        if not self.version:
            raise ValueError(
                f"version not provided and unable to find a {TF_VERSION_FILENAME} file"
            )

        # Now that a version has been selected, skip downloading if it's
        # already been downloaded
        if (self.versions_dir / str(self.version)).is_dir():
            LOGGER.verbose("Terraform version %s already installed; using it...", self.version)
            return str(self.bin)

        LOGGER.info("downloading and using Terraform version %s ...", self.version)
        download_tf_release(str(self.version), self.versions_dir, self.command_suffix)
        LOGGER.verbose("downloaded Terraform %s successfully", self.version)
        return str(self.bin)

    def list_installed(self) -> Generator[Path, None, None]:
        """List installed versions of Terraform.

        Only lists versions of Terraform that have been installed by an instance
        if this class or by tfenv.

        """
        LOGGER.verbose("checking %s for Terraform versions...", self.versions_dir)
        return self.versions_dir.rglob("*.*.*")

    def set_version(self, version: str) -> None:
        """Set current version.

        Clears cached values as needed.

        Args:
            version: Version string. Must be in the format of
                ``<major>.<minor>.<patch>`` with an optional ``-<prerelease>``.

        """
        if self.current_version == version:
            return
        self.current_version = version
        self._del_cached_property("version")

    @classmethod
    def get_version_from_executable(
        cls,
        bin_path: Path | str,
        *,
        cwd: Path | str | None = None,
        env: dict[str, str] | None = None,
    ) -> Version | None:
        """Get Terraform version from an executable.

        Args:
            bin_path: Path to the Terraform binary to retrieve the version from.
            cwd: Current working directory to use when calling the executable.
            env: Environment variable overrides.

        """
        output = subprocess.check_output([str(bin_path), "-version"], cwd=cwd, env=env).decode()
        match = re.search(cls.VERSION_OUTPUT_REGEX, output)
        if not match:
            return None
        return cls.parse_version_string(output)

    @classmethod
    def parse_version_string(cls, version: str) -> Version:
        """Parse version string into a :class:`Version`.

        Args:
            version: Version string to parse. Must be in the format of
                ``<major>.<minor>.<patch>`` with an optional ``-<prerelease>``.

        """
        match = re.search(cls.VERSION_REGEX, version)
        if not match:
            raise ValueError(f"provided version doesn't conform to regex: {cls.VERSION_REGEX}")
        return Version(match.group("version"))



"""Docker login hook."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Annotated, Any

from pydantic import ConfigDict, Field, field_validator, model_validator

from ....context import CfnginContext
from ....utils import BaseModel
from .data_models import ElasticContainerRegistry
from .hook_data import DockerHookData

if TYPE_CHECKING:
    from pydantic import ValidationInfo

LOGGER = logging.getLogger(__name__.replace("._", "."))


class LoginArgs(BaseModel):
    """Args passed to the docker.login hook."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    ctx: Annotated[CfnginContext | None, Field(alias="context", exclude=True)] = None

    dockercfg_path: str | None = None
    """Path to a non-default Docker config file."""

    ecr: ElasticContainerRegistry | None = Field(default=None, exclude=True)
    """Information describing an ECR registry."""

    email: str | None = None
    """The email for the registry account."""

    password: str
    """The plaintext password for the registry account."""

    registry: Annotated[str | None, Field(validate_default=True)] = None
    """URI of the registry to login to."""

    username: str = "AWS"
    """The registry username."""

    @model_validator(mode="before")
    @classmethod
    def _set_ecr(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Set the value of ``ecr``."""
        if "ecr" in values and isinstance(values["ecr"], dict):
            values["ecr"] = ElasticContainerRegistry.model_validate(
                {"context": values.get("context"), **values["ecr"]}
            )
        return values

    @field_validator("registry", mode="before")
    @classmethod
    def _set_registry(cls, v: Any, info: ValidationInfo) -> Any:
        """Set the value of ``registry``."""
        if v:
            return v

        ecr: ElasticContainerRegistry | None = info.data.get("ecr")
        if ecr:
            return ecr.fqn

        return None


def login(*, context: CfnginContext, **kwargs: Any) -> DockerHookData:
    """Docker login hook.

    Replicates the functionality of ``docker login`` cli command.

    kwargs are parsed by :class:`~runway.cfngin.hooks.docker.LoginArgs`.

    """
    args = LoginArgs.model_validate({"context": context, **kwargs})
    docker_hook_data = DockerHookData.from_cfngin_context(context)
    docker_hook_data.client.login(**args.model_dump())
    LOGGER.info("logged into %s", args.registry)
    return docker_hook_data.update_context(context)



"""Runway & CFNgin config model utilities."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from pathlib import Path

CFNGIN_LOOKUP_STRING_REGEX = r"^\$\{.+\}$"
RUNWAY_LOOKUP_STRING_ERROR = ValueError("field can only be a string if it's a lookup")
RUNWAY_LOOKUP_STRING_REGEX = r"^\$\{.+\}$"


def convert_null_values(v: Any) -> Any:
    """Convert a "null" string into type(None)."""
    null_strings = ["null", "none", "undefined"]
    return None if isinstance(v, str) and v.lower() in null_strings else v


def resolve_path_field(v: Path | None) -> Path | None:
    """Resolve sys_path."""
    return v.resolve() if v else v


def validate_string_is_lookup(v: Any) -> Any:
    """Validate value against regex if it's a string to ensure its a lookup."""
    if isinstance(v, str) and not re.match(RUNWAY_LOOKUP_STRING_REGEX, v):
        raise RUNWAY_LOOKUP_STRING_ERROR
    return v



"""Runway constants."""

from __future__ import annotations

from typing import Any

BOTO3_CREDENTIAL_CACHE: dict[str, Any] = {}
"""A global credential cache that can be shared among boto3 sessions.
This is inherently threadsafe thanks to the GIL.
(https://docs.python.org/3/glossary.html#term-global-interpreter-lock)
"""



"""Runway AWS Cloud Development Kit Module options."""

from __future__ import annotations

from pydantic import ConfigDict

from ...base import ConfigProperty


class RunwayCdkModuleOptionsDataModel(ConfigProperty):
    """Model for Runway AWS Cloud Development Kit Module options."""

    model_config = ConfigDict(
        extra="ignore",
        title="Runway AWS Cloud Development Kit Module options",
        validate_default=True,
        validate_assignment=True,
    )

    build_steps: list[str] = []
    skip_npm_ci: bool = False



"""Action architecture.

.. Derived from software distributed by Amazon.com, Inc - http://aws.amazon.com/apache2.0/
   https://github.com/aws/aws-cli/blob/83b43782dd/awscli/customizations/s3/subcommands.py

"""

from __future__ import annotations

import logging
from queue import Queue
from typing import TYPE_CHECKING, Any, cast

from typing_extensions import Literal, TypedDict

from ......compat import cached_property
from .comparator import Comparator
from .file_generator import FileGenerator
from .file_info_builder import FileInfoBuilder
from .filters import Filter
from .format_path import FormatPath
from .s3handler import S3TransferHandlerFactory
from .sync_strategy.base import MissingFileSync, NeverSync, SizeAndLastModifiedSync
from .transfer_config import RuntimeConfig

if TYPE_CHECKING:
    import boto3
    from botocore.session import Session
    from mypy_boto3_s3.client import S3Client

    from .format_path import FormatPathResult
    from .parameters import ParametersDataModel
    from .s3handler import S3TransferHandler
    from .sync_strategy.base import BaseSync
    from .transfer_config import TransferConfigDict

LOGGER = logging.getLogger(__name__.replace("._", "."))


_InstructionTypeDef = Literal[
    "comparator",
    "file_generator",
    "file_info_builder",
    "filters",
    "s3_handler",
    "setup",
]


class _CommandDictTypeDef(TypedDict):
    """Type definition for command_dict."""

    comparator: list[Comparator]
    file_generator: list[FileGenerator]
    file_info_builder: list[FileInfoBuilder]
    filters: list[Any]
    s3_handler: list[S3TransferHandler]
    setup: list[FormatPathResult]


class ActionArchitecture:
    """Drives the action.

    An action is performed in two steps. First a list of instructions is generated.
    This list of instructions identifies which type of components are required
    based on the name of the command and the parameters passed to the command line.
    After the instructions are generated the second step involves using the
    list of instructions to wire together an assortment of generators to
    perform the command.

    """

    def __init__(
        self,
        session: boto3.Session,
        botocore_session: Session,
        action: Literal["sync"],
        parameters: ParametersDataModel,
        runtime_config: TransferConfigDict | None = None,
    ) -> None:
        """Instantiate class."""
        self.botocore_session = botocore_session
        self.session = session
        self.action = action
        self.parameters = parameters
        self._runtime_config = runtime_config or RuntimeConfig.defaults()
        self._source_client = None

    @cached_property
    def client(self) -> S3Client:
        """Boto3 S3 client."""
        return self.session.client("s3")

    @cached_property
    def instructions(self) -> list[_InstructionTypeDef]:
        """Create the instructions based on the command name and parameters.

        Note that all commands must have an s3_handler instruction in the
        instructions and must be at the end of the instruction list because it
        sends the request to S3 and does not yield anything.

        """
        result: list[_InstructionTypeDef] = ["file_generator"]
        if self.parameters.exclude or self.parameters.include:
            result.append("filters")
        if self.action == "sync":
            result.append("comparator")
        result.append("file_info_builder")
        result.append("s3_handler")
        return result

    def choose_sync_strategies(self) -> dict[str, BaseSync]:
        """Determine the sync strategy for the command.

        It defaults to the default sync strategies but a customizable sync
        strategy can override the default strategy if it returns the instance
        of its self when the event is emitted.

        """
        sync_strategies: dict[str, BaseSync] = {
            "file_at_src_and_dest_sync_strategy": SizeAndLastModifiedSync(),
            "file_not_at_dest_sync_strategy": MissingFileSync(),
            "file_not_at_src_sync_strategy": NeverSync(),
        }

        # Determine what strategies to override if any.
        responses = cast(
            "list[tuple[Any, BaseSync]] | None",
            self.botocore_session.emit("choosing-s3-sync-strategy", params=self.parameters),
        )
        if responses is not None:
            for response in responses:
                override_sync_strategy = response[1]
                if override_sync_strategy:
                    sync_type = override_sync_strategy.sync_type
                    sync_type += "_sync_strategy"
                    sync_strategies[sync_type] = override_sync_strategy

        return sync_strategies

    def run(self) -> Literal[1, 2, 0]:
        """Wire together all of the generators and completes the action.

        First a dictionary is created that is indexed first by
        the action name. Then using the instruction, another dictionary
        can be indexed to obtain the objects corresponding to the
        particular instruction for that action. To begin the wiring,
        either a ``FileFormat`` or ``TaskInfo`` object, depending on the
        action, is put into a list. Then the function enters a while loop
        that pops off an instruction. It then determines the object needed
        and calls the call function of the object using the list as the input.
        Depending on the number of objects in the input list and the number
        of components in the list corresponding to the instruction, the call
        method of the component can be called two different ways. If the
        number of inputs is equal to the number of components a 1:1 mapping of
        inputs to components is used when calling the call function. If the
        there are more inputs than components, then a 2:1 mapping of inputs to
        components is used where the component call method takes two inputs
        instead of one. Whatever files are yielded from the call function
        is appended to a list and used as the input for the next repetition
        of the while loop until there are no more instructions.

        """
        paths_type = self.parameters.paths_type
        files = FormatPath.format(self.parameters.src, self.parameters.dest)
        rev_files = FormatPath.format(self.parameters.dest, self.parameters.src)

        action_translation = {
            "locals3": "upload",
            "s3s3": "copy",
            "s3local": "download",
            "s3": "delete",
        }
        result_queue: Queue[Any] = Queue()
        operation_name = action_translation[paths_type]

        file_generator = FileGenerator(
            client=self.client,
            operation_name=operation_name,
            follow_symlinks=self.parameters.follow_symlinks,
            page_size=self.parameters.page_size,
            result_queue=result_queue,
            request_parameters=self._get_file_generator_request_parameters_skeleton(),
        )
        rev_generator = FileGenerator(
            client=self.client,
            operation_name="",
            follow_symlinks=self.parameters.follow_symlinks,
            page_size=self.parameters.page_size,
            result_queue=result_queue,
            request_parameters=self._get_file_generator_request_parameters_skeleton(),
        )
        file_info_builder = FileInfoBuilder(client=self.client, parameters=self.parameters)
        s3_transfer_handler = S3TransferHandlerFactory(
            config_params=self.parameters, runtime_config=self._runtime_config
        )(self.client, result_queue)

        sync_strategies = self.choose_sync_strategies()

        command_dict: _CommandDictTypeDef
        if self.action == "sync":
            command_dict = {
                "setup": [files, rev_files],
                "file_generator": [file_generator, rev_generator],
                "filters": [
                    Filter.parse_params(self.parameters),
                    Filter.parse_params(self.parameters),
                ],
                "comparator": [Comparator(**sync_strategies)],
                "file_info_builder": [file_info_builder],
                "s3_handler": [s3_transfer_handler],
            }
        else:
            raise NotImplementedError("only sync is supported")

        files = command_dict["setup"]
        while self.instructions:
            instruction = self.instructions.pop(0)
            file_list = []
            components = command_dict[instruction]
            for index, comp in enumerate(components):
                if len(files) > len(components):
                    file_list.append(comp.call(*files))  # type: ignore
                else:
                    file_list.append(comp.call(files[index]))  # type: ignore
            files = file_list
        # This is kinda quirky, but each call through the instructions
        # will replaces the files attr with the return value of the
        # file_list.  The very last call is a single list of
        # [s3_handler], and the s3_handler returns the number of
        # tasks failed and the number of tasks warned.
        # This means that files[0] now contains a namedtuple with
        # the number of failed tasks and the number of warned tasks.
        # In terms of the RC, we're keeping it simple and saying
        # that > 0 failed tasks will give a 1 RC and > 0 warned
        # tasks will give a 2 RC.  Otherwise a RC of zero is returned.
        return_code: int = 0
        if files[0].num_tasks_failed > 0:  # type: ignore
            return_code = 1
        elif files[0].num_tasks_warned > 0:  # type: ignore
            return_code = 2
        return return_code

    @staticmethod
    def _get_file_generator_request_parameters_skeleton() -> dict[str, dict[str, Any]]:
        return {"HeadObject": {}, "ListObjects": {}, "ListObjectsV2": {}}



"""AWS EC2 keypair hook."""

from __future__ import annotations

import logging
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any

from botocore.exceptions import ClientError
from typing_extensions import Literal, TypedDict

from ...utils import BaseModel
from ..ui import get_raw_input

if TYPE_CHECKING:
    from mypy_boto3_ec2.client import EC2Client
    from mypy_boto3_ec2.type_defs import ImportKeyPairResultTypeDef, KeyPairTypeDef
    from mypy_boto3_ssm.client import SSMClient

    from ...context import CfnginContext

LOGGER = logging.getLogger(__name__)

KEYPAIR_LOG_MESSAGE = "keypair %s (%s) %s"


class EnsureKeypairExistsHookArgs(BaseModel):
    """Hook arguments for ``ensure_keypair_exists``."""

    keypair: str
    """Name of the key pair to ensure exists."""

    public_key_path: str | None = None
    """Path to a public key file to be imported instead of generating a new key.
    Incompatible with the SSM options, as the private key will not be available for storing.

    """

    ssm_key_id: str | None = None
    """ID of a KMS key to encrypt the SSM parameter with.
    If omitted, the default key will be used.

    """

    ssm_parameter_name: str | None = None
    """Path to an SSM store parameter to receive the generated private key
    instead of importing it or storing it locally.

    """


class KeyPairInfo(TypedDict, total=False):
    """Value returned from get_existing_key_pair."""

    file_path: Path
    fingerprint: str
    key_name: str
    status: Literal["created", "exists", "imported"]


def get_existing_key_pair(ec2: EC2Client, keypair_name: str) -> KeyPairInfo | None:
    """Get existing keypair."""
    resp = ec2.describe_key_pairs()
    keypair = next(
        (kp for kp in resp.get("KeyPairs", []) if kp.get("KeyName") == keypair_name),
        None,
    )

    if keypair:
        LOGGER.info(
            KEYPAIR_LOG_MESSAGE,
            keypair.get("KeyName"),
            keypair.get("KeyFingerprint"),
            "exists",
        )
        return {
            "status": "exists",
            "key_name": keypair.get("KeyName", ""),
            "fingerprint": keypair.get("KeyFingerprint", ""),
        }

    LOGGER.info('keypair "%s" not found', keypair_name)
    return None


def import_key_pair(
    ec2: EC2Client, keypair_name: str, public_key_data: bytes
) -> ImportKeyPairResultTypeDef:
    """Import keypair."""
    keypair = ec2.import_key_pair(
        KeyName=keypair_name, PublicKeyMaterial=public_key_data.strip(), DryRun=False
    )
    LOGGER.info(
        KEYPAIR_LOG_MESSAGE,
        keypair.get("KeyName"),
        keypair.get("KeyFingerprint"),
        "imported",
    )
    return keypair


def read_public_key_file(path: Path) -> bytes | None:
    """Read public key file."""
    try:
        data = path.read_bytes()
        if not data.startswith(b"ssh-rsa"):
            raise ValueError(
                "Bad public key data, must be an RSA key in SSH authorized "
                "keys format (beginning with `ssh-rsa`)"
            )
        return data.strip()
    except (ValueError, OSError) as err:
        LOGGER.error('failed to read public key file :%s": %s', path, str(err))
        return None


def create_key_pair_from_public_key_file(
    ec2: EC2Client, keypair_name: str, public_key_path: Path
) -> KeyPairInfo | None:
    """Create keypair from public key file."""
    public_key_data = read_public_key_file(public_key_path)
    if not public_key_data:
        return None

    keypair = import_key_pair(ec2, keypair_name, public_key_data)
    return {
        "status": "imported",
        "key_name": keypair.get("KeyName", ""),
        "fingerprint": keypair.get("KeyFingerprint", ""),
    }


def create_key_pair_in_ssm(
    ec2: EC2Client,
    ssm: SSMClient,
    keypair_name: str,
    parameter_name: str,
    kms_key_id: str | None = None,
) -> KeyPairInfo | None:
    """Create keypair in SSM."""
    keypair = create_key_pair(ec2, keypair_name)
    try:
        kms_key_label = "default"
        kms_args: dict[str, Any] = {}
        if kms_key_id:
            kms_key_label = kms_key_id
            kms_args = {"KeyId": kms_key_id}

        LOGGER.info(
            'storing generated key in SSM parameter "%s" using KMS key "%s"',
            parameter_name,
            kms_key_label,
        )

        ssm.put_parameter(
            Name=parameter_name,
            Description=f'SSH private key for KeyPair "{keypair_name}" (generated by Runway)',
            Value=keypair["KeyMaterial"],
            Type="SecureString",
            Overwrite=False,
            **kms_args,
        )
    except ClientError:
        # Erase the key pair if we failed to store it in SSM, since the
        # private key will be lost anyway

        LOGGER.exception(
            "failed to store generated key in SSM; deleting "
            "created key pair as private key will be lost"
        )
        ec2.delete_key_pair(KeyName=keypair_name, DryRun=False)
        return None

    return {
        "status": "created",
        "key_name": keypair.get("KeyName", ""),
        "fingerprint": keypair.get("KeyFingerprint", ""),
    }


def create_key_pair(ec2: EC2Client, keypair_name: str) -> KeyPairTypeDef:
    """Create keypair."""
    keypair = ec2.create_key_pair(KeyName=keypair_name, DryRun=False)
    LOGGER.info(
        KEYPAIR_LOG_MESSAGE,
        keypair.get("KeyName"),
        keypair.get("KeyFingerprint"),
        "created",
    )
    return keypair


def create_key_pair_local(ec2: EC2Client, keypair_name: str, dest_dir: Path) -> KeyPairInfo | None:
    """Create local keypair."""
    dest_dir = dest_dir.resolve()
    if not dest_dir.is_dir():
        LOGGER.error('"%s" is not a valid directory', dest_dir)
        return None

    key_path = dest_dir / f"{keypair_name}.pem"
    if key_path.is_file():
        # This mimics the old boto2 keypair.save error
        LOGGER.error('"%s" already exists in directory "%s"', key_path.name, dest_dir)
        return None

    keypair = create_key_pair(ec2, keypair_name)
    key_path.write_text(keypair.get("KeyMaterial", ""), encoding="ascii")

    return {
        "status": "created",
        "key_name": keypair.get("KeyName", ""),
        "fingerprint": keypair.get("KeyFingerprint", ""),
        "file_path": key_path,
    }


def interactive_prompt(
    keypair_name: str,
) -> tuple[Literal["create", "import"] | None, str | None]:
    """Interactive prompt."""
    if not sys.stdin.isatty():
        return None, None

    try:
        while True:
            action = get_raw_input(
                f'import or create keypair "{keypair_name}"? (import/create/cancel) '
            )

            if action.lower() == "cancel":
                break

            if action.lower() in ("i", "import"):
                path = get_raw_input("path to keypair file: ")
                return "import", path.strip()

            if action.lower() == "create":
                path = get_raw_input("directory to save keyfile: ")
                return "create", path.strip()
    except (EOFError, KeyboardInterrupt):
        return None, None

    return None, None


def ensure_keypair_exists(context: CfnginContext, *__args: Any, **kwargs: Any) -> KeyPairInfo:
    """Ensure a specific keypair exists within AWS.

    If the key doesn't exist, upload it.

    """
    args = EnsureKeypairExistsHookArgs.model_validate(kwargs)

    if args.public_key_path and args.ssm_parameter_name:
        LOGGER.error("public_key_path and ssm_parameter_name cannot be specified at the same time")
        return {}

    session = context.get_session()
    ec2 = session.client("ec2")

    keypair_info = get_existing_key_pair(ec2, args.keypair)
    if keypair_info:
        return keypair_info

    if args.public_key_path:
        keypair_info = create_key_pair_from_public_key_file(
            ec2, args.keypair, Path(args.public_key_path)
        )
    elif args.ssm_parameter_name:
        ssm = session.client("ssm")
        keypair_info = create_key_pair_in_ssm(
            ec2, ssm, args.keypair, args.ssm_parameter_name, args.ssm_key_id
        )
    else:
        action, path = interactive_prompt(args.keypair)
        if action == "import" and path:
            keypair_info = create_key_pair_from_public_key_file(ec2, args.keypair, Path(path))
        elif action == "create" and path:
            keypair_info = create_key_pair_local(ec2, args.keypair, Path(path))
        else:
            LOGGER.error("no action to find keypair or path not provided")

    if not keypair_info:
        return {}

    return keypair_info



"""Execute the AWS CLI update-kubeconfig command."""

from __future__ import annotations

import logging
import os
import subprocess
import sys
from typing import TYPE_CHECKING, Any

from runway.cfngin.lookups.handlers.output import OutputLookup
from runway.utils import which

if TYPE_CHECKING:
    from runway.context import CfnginContext

LOGGER = logging.getLogger(__name__)


def aws_eks_update_kubeconfig(context: CfnginContext, **kwargs: Any) -> bool:
    """Execute the aws cli eks update-kubeconfig command.

    Args:
        context: Context object.
        **kwargs: Arbitrary keyword arguments.

    Returns:
        boolean for whether or not the hook succeeded

    """
    if kwargs.get("cluster-name"):
        eks_cluster_name = kwargs["cluster-name"]
    else:
        eks_cluster_name = OutputLookup.handle(
            f"{kwargs['stack']}::EksClusterName", context=context
        )
    LOGGER.info("writing kubeconfig...")
    subprocess.check_output(["aws", "eks", "update-kubeconfig", "--name", eks_cluster_name])
    LOGGER.info("kubeconfig written successfully...")

    # The newly-generated kubeconfig will have introduced a dependency on the
    # awscli. This is fine if a recent version is installed, or it's invoked
    # in a virtualenv with runway
    if not os.environ.get("PIPENV_ACTIVE") and (
        not os.environ.get("VIRTUAL_ENV") and not which("aws")
    ):
        print(file=sys.stderr)  # noqa: T201
        print(  # noqa: T201
            "Warning: the generated kubeconfig uses the aws-cli for "
            "authentication, but it is not found in your environment. ",
            file=sys.stderr,
        )
    return True



"""Runway logging."""

from __future__ import annotations

import logging
from enum import IntEnum
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import MutableMapping


class LogLevels(IntEnum):
    """All available log levels."""

    NOTSET = 0
    DEBUG = 10
    VERBOSE = 15
    INFO = 20
    NOTICE = 25
    WARNING = 30
    SUCCESS = 35
    ERROR = 40
    CRITICAL = 50

    @classmethod
    def has_value(cls, value: int) -> bool:
        """Check if IntEnum has a value."""
        return value in cls._value2member_map_


# Issue with this version of LoggerAdapter https://github.com/python/typeshed/issues/7855
class PrefixAdaptor(logging.LoggerAdapter):  # type: ignore
    """LoggerAdapter that adds prefixes to messages.

    Example:
        >>> logger = PrefixAdaptor('something', logging.getLogger('example'))
        ... logger.info('my message')

    """

    def __init__(
        self,
        prefix: str,
        logger: logging.Logger,
        prefix_template: str = "{prefix}:{msg}",
    ) -> None:
        """Instantiate class.

        Args:
            prefix: Message prefix.
            logger: Logger where the prefixed messages will be sent.
            prefix_template: String that can be used with
                ".format(prefix=<prefix>, msg=<msg>)" to produce a dynamic
                message prefix.

        """
        super().__init__(logger, {})
        self.prefix = prefix
        self.prefix_template = prefix_template

    def notice(self, msg: Exception | str, *args: Any, **kwargs: Any) -> None:
        """Delegate a notice call to the underlying logger.

        Args:
            msg: String template or exception to use for the log record.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        """
        self.log(LogLevels.NOTICE, msg, *args, **kwargs)

    def process(
        self, msg: Exception | str, kwargs: MutableMapping[str, Any]
    ) -> tuple[str, MutableMapping[str, Any]]:
        """Process the message to append the prefix.

        Args:
            msg: Message to be prefixed.
            kwargs: Keyword args for the message.

        """
        return self.prefix_template.format(prefix=self.prefix, msg=msg), kwargs

    def setLevel(self, level: int | str) -> None:  # noqa: N802
        """Set the specified level on the underlying logger.

        Python 2 backport.

        """
        self.logger.setLevel(level)

    def success(self, msg: Exception | str, *args: Any, **kwargs: Any) -> None:
        """Delegate a success call to the underlying logger.

        Args:
            msg: String template or exception to use for the log record.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        """
        self.log(LogLevels.SUCCESS, msg, *args, **kwargs)

    def verbose(self, msg: Exception | str, *args: Any, **kwargs: Any) -> None:
        """Delegate a verbose call to the underlying logger.

        Args:
            msg: String template or exception to use for the log record.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        """
        self.log(LogLevels.VERBOSE, msg, *args, **kwargs)


class RunwayLogger(logging.Logger):
    """Extend built-in logger with additional levels."""

    def __init__(self, name: str, level: int | str = logging.NOTSET) -> None:
        """Instantiate the class.

        Args:
            name: Logger name.
            level: Log level.

        """
        super().__init__(name, level)
        logging.addLevelName(LogLevels.VERBOSE, LogLevels.VERBOSE.name)
        logging.addLevelName(LogLevels.NOTICE, LogLevels.NOTICE.name)
        logging.addLevelName(LogLevels.SUCCESS, LogLevels.SUCCESS.name)

    def notice(self, msg: Exception | str, *args: Any, **kwargs: Any) -> None:
        """Log 'msg % args' with severity `NOTICE`.

        Args:
            msg: String template or exception to use for the log record.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        """
        if self.isEnabledFor(LogLevels.NOTICE):
            self._log(LogLevels.NOTICE, msg, args, **kwargs)

    def success(self, msg: Exception | str, *args: Any, **kwargs: Any) -> None:
        """Log 'msg % args' with severity `SUCCESS`.

        Args:
            msg: String template or exception to use for the log record.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        """
        if self.isEnabledFor(LogLevels.SUCCESS):
            self._log(LogLevels.SUCCESS, msg, args, **kwargs)

    def verbose(self, msg: Exception | str, *args: Any, **kwargs: Any) -> None:
        """Log 'msg % args' with severity `VERBOSE`.

        Args:
            msg: String template or exception to use for the log record.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        """
        if self.isEnabledFor(LogLevels.VERBOSE):
            self._log(LogLevels.VERBOSE, msg, args, **kwargs)



"""Runway Serverless Framework Module options."""

from __future__ import annotations

from typing import Any

from pydantic import ConfigDict

from ...base import ConfigProperty


class RunwayServerlessPromotezipOptionDataModel(ConfigProperty):
    """Model for Runway Serverless module promotezip option."""

    model_config = ConfigDict(
        extra="forbid",
        title="Runway Serverless Framework Module promotezip option",
        validate_default=True,
        validate_assignment=True,
    )

    bucketname: str | None = None

    def __bool__(self) -> bool:
        """Evaluate the boolean value of the object instance."""
        return bool(self.model_dump(exclude_none=True))


class RunwayServerlessModuleOptionsDataModel(ConfigProperty):
    """Model for Runway Serverless Framework Module options."""

    model_config = ConfigDict(
        extra="ignore",
        title="Runway Serverless Framework Module options",
        validate_default=True,
        validate_assignment=True,
    )

    args: list[str] = []
    extend_serverless_yml: dict[str, Any] = {}
    promotezip: RunwayServerlessPromotezipOptionDataModel = (
        RunwayServerlessPromotezipOptionDataModel()
    )
    skip_npm_ci: bool = False



"""S3 results.

.. Derived from software distributed by Amazon.com, Inc - http://aws.amazon.com/apache2.0/
   https://github.com/aws/aws-cli/blob/83b43782dd/awscli/customizations/s3/results.py

"""

from __future__ import annotations

import logging
import queue
import sys
import threading
import time
from collections import defaultdict
from concurrent.futures import CancelledError
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    NamedTuple,
    TextIO,
    TypedDict,
    cast,
)

from s3transfer.exceptions import FatalError

from ......utils import ensure_string
from .utils import (
    OnDoneFilteredSubscriber,
    PrintTask,
    human_readable_size,
    relative_path,
    uni_print,
)

if TYPE_CHECKING:
    from types import TracebackType

    from s3transfer.futures import TransferFuture
    from typing_extensions import Literal, Self, TypeAlias

    from ......_logging import RunwayLogger
    from ......type_defs import AnyPath


LOGGER = cast("RunwayLogger", logging.getLogger(__name__.replace("._", ".")))


class CommandResult(NamedTuple):
    """Command result."""

    num_tasks_failed: int
    num_tasks_warned: int
    dest: str | None = None
    src: str | None = None
    transfer_type: str | None = None


class CtrlCResult(NamedTuple):
    """Keyboard exit."""

    exception: Exception
    dest: str | None = None
    src: str | None = None
    transfer_type: str | None = None


class DryRunResult(NamedTuple):
    """Dry run result."""

    dest: str | None = None
    src: str | None = None
    transfer_type: str | None = None


class ErrorResult(NamedTuple):
    """Error."""

    exception: BaseException
    dest: str | None = None
    src: str | None = None
    transfer_type: str | None = None


class FailureResult(NamedTuple):
    """Failure."""

    exception: Exception
    dest: str | None = None
    src: str | None = None
    transfer_type: str | None = None


class FinalTotalSubmissionsResult(NamedTuple):
    """Final total submissions."""

    total_submissions: int
    dest: str | None = None
    src: str | None = None
    transfer_type: str | None = None


class ProgressResult(NamedTuple):
    """Progress."""

    bytes_transferred: int
    timestamp: float
    total_transfer_size: int
    dest: str | None = None
    src: str | None = None
    transfer_type: str | None = None


class SuccessResult(NamedTuple):
    """Success."""

    dest: str | None = None
    src: str | None = None
    transfer_type: str | None = None


class QueuedResult(NamedTuple):
    """Queued."""

    total_transfer_size: int
    dest: str | None = None
    src: str | None = None
    transfer_type: str | None = None


AllResultTypes = (
    CommandResult,
    CtrlCResult,
    DryRunResult,
    ErrorResult,
    FailureResult,
    FinalTotalSubmissionsResult,
    ProgressResult,
    SuccessResult,
    QueuedResult,
)
AnyResultType: TypeAlias = (
    "CommandResult | CtrlCResult | DryRunResult | ErrorResult | FailureResult | FinalTotalSubmissionsResult | ProgressResult | QueuedResult | SuccessResult"  # noqa: E501
)


class _ResultHandlerMappingTypedDict(TypedDict, total=False):

    CommandResult: Callable[..., None]
    CtrlCResult: Callable[..., None]
    DryRunResult: Callable[..., None]
    ErrorResult: Callable[..., None]
    FailureResult: Callable[..., None]
    FinalTotalSubmissionsResult: Callable[..., None]
    PrintTask: Callable[..., None]
    ProgressResult: Callable[..., None]
    QueuedResult: Callable[..., None]
    SuccessResult: Callable[..., None]


class ShutdownThreadRequest:
    """Shutdown thread request."""


class BaseResultSubscriber(OnDoneFilteredSubscriber):
    """Base result subscriber."""

    TRANSFER_TYPE: ClassVar = None

    def __init__(self, result_queue: queue.Queue[Any], transfer_type: str | None = None) -> None:
        """Send result notifications during transfer process.

        Args:
            result_queue: The queue to place results to be processed later
                on.
            transfer_type: Type of transfer.

        """
        self._result_queue = result_queue
        self._result_kwargs_cache: dict[str, Any] = {}
        self._transfer_type = transfer_type
        if transfer_type is None:
            self._transfer_type = self.TRANSFER_TYPE

    def on_queued(self, future: TransferFuture, **_: Any) -> None:
        """On queue."""
        self._add_to_result_kwargs_cache(future)
        result_kwargs = self._result_kwargs_cache[future.meta.transfer_id]  # type: ignore
        queued_result = QueuedResult(**result_kwargs)
        self._result_queue.put(queued_result)

    def on_progress(self, future: TransferFuture, bytes_transferred: int, **_: Any) -> None:
        """On progress."""
        result_kwargs: dict[str, Any] = self._result_kwargs_cache.get(
            cast(str, future.meta.transfer_id), cast("dict[str, Any]", {})
        )
        progress_result = ProgressResult(
            bytes_transferred=bytes_transferred, timestamp=time.time(), **result_kwargs
        )
        self._result_queue.put(progress_result)

    def _on_success(self, future: TransferFuture) -> None:
        """On success."""
        result_kwargs = self._on_done_pop_from_result_kwargs_cache(future)
        self._result_queue.put(SuccessResult(**result_kwargs))

    def _on_failure(self, future: TransferFuture, exception: Exception) -> None:
        """On failure."""
        result_kwargs = self._on_done_pop_from_result_kwargs_cache(future)
        if isinstance(exception, CancelledError):
            error_result_cls = CtrlCResult
            if isinstance(exception, FatalError):
                error_result_cls = ErrorResult
            self._result_queue.put(error_result_cls(exception=exception))
        else:
            self._result_queue.put(FailureResult(exception=exception, **result_kwargs))

    def _add_to_result_kwargs_cache(self, future: TransferFuture) -> None:
        """Add to results cache."""
        src, dest = self._get_src_dest(future)
        result_kwargs = {
            "transfer_type": self._transfer_type,
            "src": src,
            "dest": dest,
            "total_transfer_size": future.meta.size,
        }
        self._result_kwargs_cache[cast(str, future.meta.transfer_id)] = result_kwargs

    def _on_done_pop_from_result_kwargs_cache(self, future: TransferFuture) -> dict[str, Any]:
        """On done, pop from results cache."""
        result_kwargs: dict[str, Any] = self._result_kwargs_cache.pop(
            cast(str, future.meta.transfer_id)
        )
        result_kwargs.pop("total_transfer_size")
        return result_kwargs

    def _get_src_dest(self, future: TransferFuture) -> tuple[str, str]:
        """Get source destination."""
        raise NotImplementedError("_get_src_dest()")


class UploadResultSubscriber(BaseResultSubscriber):
    """Upload result subscriber."""

    TRANSFER_TYPE: ClassVar[Literal["upload"]] = "upload"

    def _get_src_dest(self, future: TransferFuture) -> tuple[str, str]:
        call_args = future.meta.call_args
        src = self._get_src(call_args.fileobj)
        dest = "s3://" + call_args.bucket + "/" + call_args.key
        return src, dest

    def _get_src(self, fileobj: AnyPath) -> str:
        return relative_path(fileobj)


class UploadStreamResultSubscriber(UploadResultSubscriber):
    """Upload stream result subscriber."""

    def _get_src(self, fileobj: AnyPath) -> str:  # noqa: ARG002
        return "-"


class DownloadResultSubscriber(BaseResultSubscriber):
    """Download result subscriber."""

    TRANSFER_TYPE: ClassVar[Literal["download"]] = "download"

    def _get_src_dest(self, future: TransferFuture) -> tuple[str, str]:
        call_args = future.meta.call_args
        src = "s3://" + call_args.bucket + "/" + call_args.key
        dest = self._get_dest(call_args.fileobj)
        return src, dest

    def _get_dest(self, fileobj: AnyPath) -> str:
        return relative_path(fileobj)


class DownloadStreamResultSubscriber(DownloadResultSubscriber):
    """Download stream result subscriber."""

    def _get_dest(self, fileobj: AnyPath) -> str:  # noqa: ARG002
        return "-"


class CopyResultSubscriber(BaseResultSubscriber):
    """Copy result subscriber."""

    TRANSFER_TYPE: ClassVar[Literal["copy"]] = "copy"

    def _get_src_dest(self, future: TransferFuture) -> tuple[str, str]:
        call_args = future.meta.call_args
        copy_source = call_args.copy_source
        src = "s3://" + copy_source["Bucket"] + "/" + copy_source["Key"]
        dest = "s3://" + call_args.bucket + "/" + call_args.key
        return src, dest


class DeleteResultSubscriber(BaseResultSubscriber):
    """Delete result subscriber."""

    TRANSFER_TYPE: ClassVar[Literal["delete"]] = "delete"

    def _get_src_dest(self, future: TransferFuture) -> tuple[str, None]:  # type: ignore
        call_args = future.meta.call_args
        src = "s3://" + call_args.bucket + "/" + call_args.key
        return src, None


class BaseResultHandler:
    """Base handler class to be called in the ResultProcessor."""

    def __call__(self, result: Any) -> None:
        """Call instance of class."""
        raise NotImplementedError("__call__()")


class ResultRecorder(BaseResultHandler):
    """Record and track transfer statistics based on results received."""

    def __init__(self) -> None:
        """Instantiate class."""
        self.bytes_transferred = 0
        self.bytes_failed_to_transfer = 0
        self.files_transferred = 0
        self.files_failed = 0
        self.files_warned = 0
        self.errors = 0
        self.expected_bytes_transferred = 0
        self.expected_files_transferred = 0
        self.final_expected_files_transferred = None

        self.start_time = None
        self.bytes_transfer_speed = 0

        self._ongoing_progress = defaultdict(int)
        self._ongoing_total_sizes: dict[str, int] = {}

        self._result_handler_map: _ResultHandlerMappingTypedDict = {
            "CtrlCResult": self._record_error_result,
            "ErrorResult": self._record_error_result,
            "FailureResult": self._record_failure_result,
            "FinalTotalSubmissionsResult": self._record_final_expected_files,
            "PrintTask": self._record_warning_result,
            "ProgressResult": self._record_progress_result,
            "QueuedResult": self._record_queued_result,
            "SuccessResult": self._record_success_result,
        }

    def expected_totals_are_final(self) -> bool:
        """Assess if expected totals are final."""
        return self.final_expected_files_transferred == self.expected_files_transferred

    def __call__(self, result: AnyResultType | PrintTask) -> None:
        """Record the result of an individual Result object."""
        self._result_handler_map.get(type(result).__name__, self._record_noop)(result=result)

    @staticmethod
    def _get_ongoing_dict_key(result: AnyResultType | object) -> str:
        if not isinstance(result, AllResultTypes):
            raise TypeError(
                "Any result using _get_ongoing_dict_key must be one of "
                f"{', '.join(str(i) for i in AllResultTypes)}. "
                f"Provided result is of type: {type(result)}"
            )
        key_parts: list[str] = []
        for result_property in [result.transfer_type, result.src, result.dest]:
            if result_property is not None:
                key_parts.append(ensure_string(result_property))  # noqa: PERF401
        return ":".join(key_parts)

    def _pop_result_from_ongoing_dicts(self, result: AnyResultType) -> tuple[int, int | None]:
        ongoing_key = self._get_ongoing_dict_key(result)
        total_progress = self._ongoing_progress.pop(ongoing_key, 0)
        total_file_size = self._ongoing_total_sizes.pop(ongoing_key, None)
        return total_progress, total_file_size

    def _record_noop(self, **_: Any) -> None:
        """If result does not have a handler, then do nothing with it."""

    def _record_queued_result(self, result: QueuedResult, **_: Any) -> None:
        if self.start_time is None:
            self.start_time = time.time()
        total_transfer_size = result.total_transfer_size
        self._ongoing_total_sizes[self._get_ongoing_dict_key(result)] = total_transfer_size
        # The total transfer size can be None if we do not know the size
        # immediately so do not add to the total right away.
        if total_transfer_size:
            self.expected_bytes_transferred += total_transfer_size
        self.expected_files_transferred += 1

    def _record_progress_result(self, result: ProgressResult, **_: Any) -> None:
        if self.start_time is None:
            self.start_time = time.time()
        bytes_transferred = result.bytes_transferred
        self._update_ongoing_transfer_size_if_unknown(result)
        self._ongoing_progress[self._get_ongoing_dict_key(result)] += bytes_transferred
        self.bytes_transferred += bytes_transferred
        # Since the start time is captured in the result recorder and
        # capture timestamps in the subscriber, there is a chance that if
        # a progress result gets created right after the queued result
        # gets created that the timestamp on the progress result is less
        # than the timestamp of when the result processor actually
        # processes that initial queued result. So this will avoid
        # negative progress being displayed or zero division occurring.
        if result.timestamp > self.start_time:
            self.bytes_transfer_speed = self.bytes_transferred / (
                result.timestamp - self.start_time
            )

    def _update_ongoing_transfer_size_if_unknown(self, result: ProgressResult) -> None:
        """Handle transfer size unknown but shown in progress result."""
        ongoing_key = self._get_ongoing_dict_key(result)
        if self._ongoing_total_sizes.get(ongoing_key) is None:
            total_transfer_size = result.total_transfer_size
            # If the total size is no longer None that means we just learned
            # of the size so let's update the appropriate places with this
            # knowledge
            if (
                result.total_transfer_size is not None
            ):  # pyright: ignore[reportUnnecessaryComparison]
                self._ongoing_total_sizes[ongoing_key] = total_transfer_size
                # Figure out how many bytes have been unaccounted for as
                # the recorder has been keeping track of how many bytes
                # it has seen so far and add it to the total expected amount.
                ongoing_progress = self._ongoing_progress[ongoing_key]
                unaccounted_bytes = total_transfer_size - ongoing_progress
                self.expected_bytes_transferred += unaccounted_bytes
            # If we still do not know what the total transfer size is
            # just update the expected bytes with the know bytes transferred
            # as we know at the very least, those bytes are expected.
            else:
                self.expected_bytes_transferred += result.bytes_transferred

    def _record_success_result(self, result: AnyResultType, **_: Any) -> None:
        self._pop_result_from_ongoing_dicts(result)
        self.files_transferred += 1

    def _record_failure_result(self, result: AnyResultType, **_: Any) -> None:
        """On failure, account for the failure in count for bytes transferred."""
        total_progress, total_file_size = self._pop_result_from_ongoing_dicts(result)
        if total_file_size is not None:
            progress_left = total_file_size - total_progress
            self.bytes_failed_to_transfer += progress_left

        self.files_failed += 1
        self.files_transferred += 1

    def _record_warning_result(self, **_: Any) -> None:
        self.files_warned += 1

    def _record_error_result(self, **_: Any) -> None:
        self.errors += 1

    def _record_final_expected_files(self, result: FinalTotalSubmissionsResult, **_: Any) -> None:
        self.final_expected_files_transferred = result.total_submissions


class ResultPrinter(BaseResultHandler):
    """Prints status of ongoing transfer."""

    _FILES_REMAINING: ClassVar[str] = "{remaining_files} file(s) remaining"
    _ESTIMATED_EXPECTED_TOTAL: ClassVar[str] = "~{expected_total}"
    _STILL_CALCULATING_TOTALS: ClassVar[str] = " (calculating...)"
    BYTE_PROGRESS_FORMAT: ClassVar[str] = (
        "Completed {bytes_completed}/{expected_bytes_completed} "
        "({transfer_speed}) with " + _FILES_REMAINING
    )
    FILE_PROGRESS_FORMAT: ClassVar[str] = (
        "Completed {files_completed} file(s) with " + _FILES_REMAINING
    )
    SUCCESS_FORMAT: ClassVar[str] = "{transfer_type}: {transfer_location}"
    DRY_RUN_FORMAT: ClassVar[str] = "(dryrun) " + SUCCESS_FORMAT
    FAILURE_FORMAT: ClassVar[str] = "{transfer_type} failed: {transfer_location} {exception}"
    WARNING_FORMAT: ClassVar[str] = "{message}"
    ERROR_FORMAT: ClassVar[str] = "fatal error: {exception}"
    CTRL_C_MSG: ClassVar[str] = "cancelled: ctrl-c received"

    SRC_DEST_TRANSFER_LOCATION_FORMAT: ClassVar[str] = "{src} to {dest}"
    SRC_TRANSFER_LOCATION_FORMAT: ClassVar[str] = "{src}"

    def __init__(
        self,
        result_recorder: ResultRecorder,
        *,
        out_file: TextIO | None = None,
        error_file: TextIO | None = None,
    ) -> None:
        """Instantiate class.

        Args:
            result_recorder: The associated result recorder
            out_file: Location to write progress and success statements.
                By default, the location is sys.stdout.
            error_file: Location to write warnings and errors.
                By default, the location is sys.stderr.

        """
        self._result_recorder = result_recorder
        self._out_file = out_file
        if self._out_file is None:
            self._out_file = sys.stdout
        self._error_file = error_file
        if self._error_file is None:
            self._error_file = sys.stderr
        self._progress_length = 0
        self._result_handler_map: _ResultHandlerMappingTypedDict = {
            "CtrlCResult": self._print_ctrl_c,
            "DryRunResult": self._print_dry_run,
            "ErrorResult": self._print_error,
            "FailureResult": self._print_failure,
            "FinalTotalSubmissionsResult": self._clear_progress_if_no_more_expected_transfers,
            "PrintTask": self._print_warning,
            "ProgressResult": self._print_progress,
            "SuccessResult": self._print_success,
        }

    def __call__(self, result: AnyResultType | PrintTask) -> None:
        """Print the progress of the ongoing transfer based on a result."""
        self._result_handler_map.get(type(result).__name__, self._print_noop)(result=result)

    def _print_noop(self, **_: Any) -> None:
        """If result does not have a handler, then do nothing with it."""

    def _print_dry_run(self, result: DryRunResult, **_: Any) -> None:
        statement = self.DRY_RUN_FORMAT.format(
            transfer_type=result.transfer_type,
            transfer_location=self._get_transfer_location(result),
        )
        LOGGER.notice(statement)

    def _print_success(self, result: SuccessResult, **_: Any) -> None:
        success_statement = self.SUCCESS_FORMAT.format(
            transfer_type=result.transfer_type,
            transfer_location=self._get_transfer_location(result),
        )
        LOGGER.notice(success_statement)
        self._redisplay_progress()

    def _print_failure(self, result: FailureResult, **_: Any) -> None:
        failure_statement = self.FAILURE_FORMAT.format(
            transfer_type=result.transfer_type,
            transfer_location=self._get_transfer_location(result),
            exception=result.exception,
        )
        LOGGER.error(failure_statement)
        self._redisplay_progress()

    def _print_warning(self, result: Any, **_: Any) -> None:
        warning_statement = self.WARNING_FORMAT.format(message=result.message)
        LOGGER.warning(warning_statement)
        self._redisplay_progress()

    def _print_error(self, result: ErrorResult, **_: Any) -> None:
        LOGGER.error(self.ERROR_FORMAT.format(exception=result.exception))

    def _print_ctrl_c(self, result: CtrlCResult, **_: Any) -> None:  # noqa: ARG002
        LOGGER.warning(self.CTRL_C_MSG)

    def _get_transfer_location(self, result: AnyResultType) -> str:
        if result.dest is None:
            return self.SRC_TRANSFER_LOCATION_FORMAT.format(src=result.src)
        return self.SRC_DEST_TRANSFER_LOCATION_FORMAT.format(src=result.src, dest=result.dest)

    def _redisplay_progress(self) -> None:
        # Reset to zero because done statements are printed with new lines
        # meaning there are no carriage returns to take into account when
        # printing the next line.
        self._progress_length = 0
        self._add_progress_if_needed()

    def _add_progress_if_needed(self) -> None:
        if self._has_remaining_progress():
            self._print_progress()

    def _print_progress(self, **_: Any) -> None:
        # Get all of the statistics in the correct form.
        remaining_files = self._get_expected_total(
            str(
                self._result_recorder.expected_files_transferred
                - self._result_recorder.files_transferred
            )
        )

        # Create the display statement.
        if self._result_recorder.expected_bytes_transferred > 0:
            bytes_completed = human_readable_size(
                self._result_recorder.bytes_transferred
                + self._result_recorder.bytes_failed_to_transfer
            )
            expected_bytes_completed = self._get_expected_total(
                human_readable_size(self._result_recorder.expected_bytes_transferred)
            )

            transfer_speed = (
                human_readable_size(self._result_recorder.bytes_transfer_speed) or "0 Bytes"
            ) + "/s"
            progress_statement = self.BYTE_PROGRESS_FORMAT.format(
                bytes_completed=bytes_completed,
                expected_bytes_completed=expected_bytes_completed,
                transfer_speed=transfer_speed,
                remaining_files=remaining_files,
            )
        else:
            # We're not expecting any bytes to be transferred, so we should
            # only print of information about number of files transferred.
            progress_statement = self.FILE_PROGRESS_FORMAT.format(
                files_completed=self._result_recorder.files_transferred,
                remaining_files=remaining_files,
            )

        if not self._result_recorder.expected_totals_are_final():
            progress_statement += self._STILL_CALCULATING_TOTALS

        # Make sure that it overrides any previous progress bar.
        progress_statement = self._adjust_statement_padding(progress_statement, ending_char="\r")
        # We do not want to include the carriage return in this calculation
        # as progress length is used for determining whitespace padding.
        # So we subtract one off of the length.
        self._progress_length = len(progress_statement) - 1

        # Print the progress out.
        self._print_to_out_file(progress_statement)

    def _get_expected_total(self, expected_total: str | None) -> str | None:
        if not self._result_recorder.expected_totals_are_final():
            return self._ESTIMATED_EXPECTED_TOTAL.format(expected_total=expected_total)
        return expected_total

    def _adjust_statement_padding(self, print_statement: str, ending_char: str = "\n") -> str:
        print_statement = print_statement.ljust(self._progress_length, " ")
        return print_statement + ending_char

    def _has_remaining_progress(self) -> bool:
        if not self._result_recorder.expected_totals_are_final():
            return True
        actual = self._result_recorder.files_transferred
        expected = self._result_recorder.expected_files_transferred
        return actual != expected

    def _print_to_out_file(self, statement: str) -> None:
        uni_print(statement, self._out_file)

    def _clear_progress_if_no_more_expected_transfers(self, **_: Any) -> None:
        if self._progress_length and not self._has_remaining_progress():
            uni_print(self._adjust_statement_padding(""), self._out_file)


class NoProgressResultPrinter(ResultPrinter):
    """A result printer that doesn't print progress."""

    def _print_progress(self, **_: Any) -> None:
        pass


class OnlyShowErrorsResultPrinter(ResultPrinter):
    """A result printer that only prints out errors."""

    def _print_progress(self, **_: Any) -> None:
        pass

    def _print_success(self, result: Any, **_: Any) -> None:
        pass


class ResultProcessor(threading.Thread):
    """Thread to process results from result queue.

    This includes recording statistics and printing transfer status

    """

    def __init__(
        self,
        result_queue: queue.Queue[Any],
        result_handlers: list[Callable[..., Any]] | None = None,
    ) -> None:
        """Instantiate class.

        Args:
            result_queue: The result queue to process results from
            result_handlers: A list of callables that take a result in as
                a parameter to process the result for that handler.

        """
        threading.Thread.__init__(self)
        self._result_queue = result_queue
        self._result_handlers = result_handlers or []
        self._result_handlers_enabled = True

    def run(self) -> None:
        """Run."""
        while True:
            try:
                result = self._result_queue.get(True)
                if isinstance(result, ShutdownThreadRequest):
                    LOGGER.debug(
                        "Shutdown request received in result processing "
                        "thread, shutting down result thread."
                    )
                    break
                if self._result_handlers_enabled:
                    self._process_result(result)
                # ErrorResults are fatal to the command. If a fatal error
                # is seen, we know that the command is trying to shutdown
                # so disable all of the handlers and quickly consume all
                # of the results in the result queue in order to get to
                # the shutdown request to clean up the process.
                if isinstance(result, ErrorResult):
                    self._result_handlers_enabled = False
            except queue.Empty:  # cov: ignore
                pass

    def _process_result(self, result: AnyResultType) -> None:
        for result_handler in self._result_handlers:
            try:
                result_handler(result)
            except Exception as exc:  # noqa: BLE001
                LOGGER.debug(
                    "Error processing result %s with handler %s: %s",
                    result,
                    result_handler,
                    exc,
                    exc_info=True,
                )


class CommandResultRecorder:
    """Records the result for an entire command.

    It will fully process all results in a result queue and determine
    a CommandResult representing the entire command.

    """

    def __init__(
        self,
        result_queue: queue.Queue[Any],
        result_recorder: ResultRecorder,
        result_processor: ResultProcessor,
    ) -> None:
        """Instantiate class.

        Args:
            result_queue: The result queue in which results are placed on
                and processed from.
            result_recorder: The result recorder to track the various
                results sent through the result queue
            result_processor: The result processor to process results
                placed on the queue

        """
        self.result_queue = result_queue
        self._result_recorder = result_recorder
        self._result_processor = result_processor

    def start(self) -> None:
        """Start."""
        self._result_processor.start()

    def shutdown(self) -> None:
        """Shutdown."""
        self.result_queue.put(ShutdownThreadRequest())
        self._result_processor.join()

    def get_command_result(self) -> CommandResult:
        """Get the CommandResult representing the result of a command."""
        return CommandResult(
            num_tasks_failed=self._result_recorder.files_failed + self._result_recorder.errors,
            num_tasks_warned=self._result_recorder.files_warned,
        )

    def notify_total_submissions(self, total: int) -> None:
        """Notify total submissions."""
        self.result_queue.put(FinalTotalSubmissionsResult(total_submissions=total))

    def __enter__(self) -> Self:
        """Enter the context manager.

        Returns:
            Instance of the context manager.

        """
        self.start()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None:
        """Exit the context manager."""
        if exc_type:
            LOGGER.debug(
                "Exception caught during command execution: %s",
                exc_value,
                exc_info=True,
            )
            if exc_value:
                self.result_queue.put(ErrorResult(exception=exc_value))
            self.shutdown()
            return True  # suppress error as it has been handled by the context manager
        self.shutdown()
        return None



"""Runway module object."""

from __future__ import annotations

import concurrent.futures
import json
import logging
import multiprocessing
import sys
from typing import TYPE_CHECKING, Any, cast

import yaml

from ..._logging import PrefixAdaptor
from ...compat import cached_property
from ...config.components.runway import RunwayVariablesDefinition
from ...config.models.runway import (
    RunwayEnvVarsType,
    RunwayFutureDefinitionModel,
    RunwayVariablesDefinitionModel,
)
from ...utils import change_dir, flatten_path_lists, merge_dicts
from ..providers import aws
from ._module_path import ModulePath
from ._module_type import RunwayModuleType

if TYPE_CHECKING:
    from ..._logging import RunwayLogger
    from ...config.components.runway import (
        RunwayDeploymentDefinition,
        RunwayModuleDefinition,
    )
    from ...config.models.runway import RunwayEnvironmentsType
    from ...context import RunwayContext
    from ..type_defs import RunwayActionTypeDef

LOGGER = cast("RunwayLogger", logging.getLogger(__name__.replace("._", ".")))


class Module:
    """Runway module."""

    ctx: RunwayContext
    logger: PrefixAdaptor
    name: str

    def __init__(
        self,
        context: RunwayContext,
        definition: RunwayModuleDefinition,
        deployment: RunwayDeploymentDefinition = None,
        future: RunwayFutureDefinitionModel = None,
        variables: RunwayVariablesDefinition = None,
    ) -> None:
        """Instantiate class.

        Args:
            context: Runway context object.
            definition: A single module definition.
            deployment: Deployment that this
                module is a part of.
            future: Future functionality
                configuration.
            variables: Runway variables.

        """
        self.__deployment = deployment
        self.__future = future or RunwayFutureDefinitionModel()
        self.__variables = variables or RunwayVariablesDefinition(RunwayVariablesDefinitionModel())
        self.ctx = context.copy()  # each module has it's own instance of context
        definition.resolve(self.ctx, variables=variables)
        self.definition = definition
        self.name = self.definition.name
        self.logger = PrefixAdaptor(self.fqn, LOGGER)

    @cached_property
    def child_modules(self) -> list[Module]:
        """Return child modules."""
        return [
            self.__class__(
                context=self.ctx,
                definition=child,
                deployment=self.__deployment,
                future=self.__future,
                variables=self.__variables,
            )
            for child in self.definition.child_modules
        ]

    @cached_property
    def environment_matches_defined(self) -> bool | None:
        """Environment matches one of the defined environments.

        Will return None if there is nothing defined for the current environment.

        """
        return validate_environment(self.ctx, self.environments, logger=self.logger)

    @cached_property
    def environments(self) -> RunwayEnvironmentsType:
        """Environments defined for the deployment and module."""
        tmp: RunwayEnvironmentsType = self.definition.environments
        if self.__deployment:
            tmp = merge_dicts(self.__deployment.environments, tmp)
        if self.opts_from_file:
            tmp = merge_dicts(tmp, self.opts_from_file.get("environments", {}))
        return tmp

    @cached_property
    def fqn(self) -> str:
        """Fully qualified name."""
        if not self.__deployment:
            return self.name
        return f"{self.__deployment.name}.{self.name}"

    @cached_property
    def opts_from_file(self) -> dict[str, Any]:
        """Load module options from local file."""
        opts_file = self.path.module_root / "runway.module.yml"
        if opts_file.is_file():
            self.logger.verbose("module-level config file found")
            return yaml.safe_load(opts_file.read_text())
        return {}

    @cached_property
    def path(self) -> ModulePath:  # lazy load the path
        """Return resolve module path."""
        return ModulePath.parse_obj(
            self.definition,
            cache_dir=self.ctx.work_dir / "cache",
            deploy_environment=self.ctx.env,
        )

    @cached_property
    def payload(self) -> dict[str, Any]:  # lazy load the payload
        """Return payload to be passed to module class handler class."""
        payload: dict[str, Any] = {}
        if self.__deployment:
            payload.update(
                {
                    "options": self.__deployment.module_options,
                    "parameters": self.__deployment.parameters,
                }
            )
        payload = merge_dicts(payload, self.definition.data)
        payload = merge_dicts(payload, self.opts_from_file)
        payload["explicitly_enabled"] = bool(self.environment_matches_defined)
        self.__merge_env_vars(payload.pop("env_vars", {}))
        return payload

    @cached_property
    def should_skip(self) -> bool:
        """Whether the module should be skipped by Runway."""
        if isinstance(self.environment_matches_defined, bool):
            return not self.environment_matches_defined
        return False

    @cached_property
    def type(self) -> RunwayModuleType:
        """Determine Runway module type."""
        return RunwayModuleType(
            path=self.path.module_root,
            class_path=self.definition.class_path,
            type_str=self.definition.type,
        )

    @cached_property
    def use_async(self) -> bool:
        """Whether to use asynchronous method."""
        return bool(self.definition.child_modules and self.ctx.use_concurrent)

    def deploy(self) -> None:
        """Deploy the module.

        High level method for running a module.

        """
        if not self.child_modules:
            return self.run("deploy")
        if self.use_async:
            return self.__async("deploy")
        return self.__sync("deploy")

    def destroy(self) -> None:
        """Destroy the module.

        High level method for running a module.

        """
        if not self.child_modules:
            return self.run("destroy")
        if self.use_async:
            return self.__async("destroy")
        return self.__sync("destroy")

    def init(self) -> None:
        """Initialize/bootstrap module.

        High level method for running a deployment.

        """
        if not self.child_modules:
            return self.run("init")
        if self.use_async:
            return self.__async("init")
        return self.__sync("init")

    def plan(self) -> None:
        """Plan for the next deploy of the module.

        High level method for running a module.

        """
        if not self.child_modules:
            return self.run("plan")
        if self.use_async:
            self.logger.info("processing of modules will be done in parallel during deploy/destroy")
        return self.__sync("plan")

    def run(self, action: RunwayActionTypeDef) -> None:
        """Run a single module.

        Low level API access to run a module object.

        Args:
            action: Name of action to run.

        """
        LOGGER.info("")
        self.logger.notice("processing module in %s (in progress)", self.ctx.env.aws_region)
        self.logger.verbose("module payload: %s", json.dumps(self.payload))
        if self.should_skip:
            return
        with change_dir(self.path.module_root):
            # dynamically load the particular module's class, 'get' the method
            # associated with the command, and call the method.
            inst = self.type.module_class(
                self.ctx, module_root=self.path.module_root, **self.payload
            )
            if hasattr(inst, action):
                inst[action]()
            else:
                self.logger.error('"%s" is missing method "%s"', inst, action)
                sys.exit(1)
        self.logger.success("processing module in %s (complete)", self.ctx.env.aws_region)

    def __async(self, action: RunwayActionTypeDef) -> None:
        """Execute asynchronously.

        Args:
            action: Name of action to run.

        """
        self.logger.info("processing modules in parallel... (output will be interwoven)")
        # Can't use threading or ThreadPoolExecutor here because
        # we need to be able to do things like `cd` which is not
        # thread safe.
        with concurrent.futures.ProcessPoolExecutor(
            max_workers=self.ctx.env.max_concurrent_modules,
            mp_context=multiprocessing.get_context("fork"),
        ) as executor:
            futures = [executor.submit(child.run, action) for child in self.child_modules]
        for job in futures:
            job.result()  # raise exceptions / exit as needed

    def __sync(self, action: RunwayActionTypeDef) -> None:
        """Execute synchronously.

        Args:
            action: Name of action to run.

        """
        self.logger.info("processing modules sequentially...")
        for module in self.child_modules:
            module.run(action)

    def __merge_env_vars(self, env_vars: RunwayEnvVarsType) -> None:
        """Merge defined env_vars into context.env_vars."""
        if env_vars:
            resolved_env_vars = flatten_path_lists(env_vars, str(self.ctx.env.root_dir))
            if resolved_env_vars:
                self.logger.verbose(
                    "environment variable overrides are being applied to this module"
                )
                self.logger.debug("environment variable overrides: %s", resolved_env_vars)
                self.ctx.env.vars = merge_dicts(self.ctx.env.vars, resolved_env_vars)

    @classmethod
    def run_list(
        cls,
        action: RunwayActionTypeDef,
        context: RunwayContext,
        modules: list[RunwayModuleDefinition],
        variables: RunwayVariablesDefinition,
        deployment: RunwayDeploymentDefinition = None,
        future: RunwayFutureDefinitionModel | None = None,
    ) -> None:
        """Run a list of modules.

        Args:
            action: Name of action to run.
            context: Runway context.
            modules: List of modules to run.
            variables: Variable definition for resolving lookups in the module.
            deployment: Deployment the modules are a part of.
            future: Future functionality configuration.

        """
        for module in modules:
            cls(
                context=context,
                definition=module,
                deployment=deployment,
                future=future,
                variables=variables,
            )[action]()

    def __getitem__(self, key: str) -> Any:
        """Make the object subscriptable.

        Args:
            key: Attribute to get.

        """
        return getattr(self, key)


def validate_environment(
    context: RunwayContext,
    env_def: bool | dict[str, Any] | int | str | list[str] | None,
    logger: PrefixAdaptor | RunwayLogger = LOGGER,
) -> bool | None:
    """Check if an environment should be deployed to.

    Args:
        context: Runway context object.
        env_def: Runway module definition.
        logger: Logger to log messages to.

    Returns:
        Boolean value of whether to deploy or not.

    """
    if isinstance(env_def, bool) or not env_def:
        if env_def is True:
            logger.verbose("explicitly enabled")
        elif env_def is False:
            logger.info("skipped; explicitly disabled")
        else:
            logger.verbose("environment not defined; module will determine deployment")
            env_def = None
        return env_def
    if isinstance(env_def, dict):
        if context.env.name not in env_def:
            logger.info("skipped; environment not in definition")
            return False
        return validate_environment(context, env_def.get(context.env.name, False), logger=logger)

    account = aws.AccountDetails(context)
    accepted_values = [
        f"{account.id}/{context.env.aws_region}",
        account.id,
        context.env.aws_region,
        int(account.id),
    ]
    result = False

    if isinstance(env_def, (int, str)):
        logger.debug('checking if "%s" in %s', env_def, accepted_values)
        result = env_def in accepted_values
    elif isinstance(env_def, list):  # type: ignore
        logger.debug("checking if any(%s in %s)", env_def, accepted_values)
        result = any(val in env_def for val in accepted_values)
    else:
        logger.warning('skipped; unsupported type for environments "%s"', type(env_def))
        return False

    if not result:
        logger.info("skipped; account_id/region mismatch")
    return result



"""Response data models."""

from __future__ import annotations

from typing import Annotated

from pydantic import ConfigDict, Field

from runway.utils import BaseModel


class AwsLambdaHookDeployResponse(BaseModel):
    """Data model for AwsLambdaHook deploy response.

    When returned by the hook as ``hook_data``, this model is dumped to a
    standard :class:`~typing.Dict` using the field's aliases as the ``key`` in
    place of the attribute names.
    This is done so that the ``key`` is a direct match to a CloudFormation
    Property where the value should be used.

    """

    model_config = ConfigDict(extra="forbid", populate_by_name=True)

    bucket_name: Annotated[str, Field(alias="S3Bucket")]
    """Name of the S3 Bucket where the deployment package is located. (alias ``S3Bucket``)"""

    code_sha256: Annotated[str, Field(alias="CodeSha256")]
    """SHA256 of the deployment package.
    This can be used by CloudFormation as the value of ``AWS::Lambda::Version.CodeSha256``.
    (alias ``CodeSha256``)

    """

    compatible_architectures: Annotated[
        list[str] | None, Field(alias="CompatibleArchitectures")
    ] = None
    """A list of compatible instruction set architectures.
    (https://docs.aws.amazon.com/lambda/latest/dg/foundation-arch.html)
    (alias ``CompatibleArchitectures``)

    """

    compatible_runtimes: Annotated[list[str] | None, Field(alias="CompatibleRuntimes")] = None
    """A list of compatible function runtimes.
    Used for filtering with ``ListLayers`` and ``ListLayerVersions``.
    (alias ``CompatibleRuntimes``)

    """

    license: Annotated[str | None, Field(alias="License")] = None
    """The layer's software license (alias ``License``). Can be any of the following:

    - A SPDX license identifier (e.g. ``MIT``).
    - The URL of a license hosted on the internet (e.g.
      ``https://opensource.org/licenses/MIT``).
    - The full text of the license.

    """

    object_key: Annotated[str, Field(alias="S3Key")]
    """Key (file path) of the deployment package S3 Object. (alias ``S3Key``)"""

    object_version_id: Annotated[str | None, Field(alias="S3ObjectVersion")] = None
    """The version ID of the deployment package S3 Object.
    This will only have a value if the S3 Bucket has versioning enabled.
    (alias ``S3ObjectVersion``)

    """

    runtime: Annotated[str, Field(alias="Runtime")]
    """Runtime of the Lambda Function. (alias ``Runtime``)"""



"""Runway ``variables`` definition model."""

from __future__ import annotations

from pathlib import Path
from typing import Annotated

from pydantic import ConfigDict, Field, field_validator

from .. import utils
from ..base import ConfigProperty


class RunwayVariablesDefinitionModel(ConfigProperty):
    """A variable definitions for the Runway config file.

    This is used to resolve the ``var`` lookup.

    """

    model_config = ConfigDict(
        extra="allow",
        json_schema_extra={
            "description": "A variable definitions for the Runway config file. "
            "This is used to resolve the 'var' lookup.",
        },
        title="Runway Variables Definition",
        validate_default=True,
        validate_assignment=True,
    )

    file_path: Annotated[
        Path | None,
        Field(
            title="Variables File Path",
            description="Explicit path to a variables file that will be loaded and "
            "merged with the variables defined here.",
        ),
    ] = None
    """Explicit path to a variables file that will be loaded and merged with the variables defined here."""

    sys_path: Annotated[
        Path,
        Field(
            description="Directory to use as the root of a relative 'file_path'. "
            "If not provided, the current working directory is used.",
        ),
    ] = "./"  # pyright: ignore[reportAssignmentType]
    """Directory to use as the root of a relative 'file_path'.
    If not provided, the current working directory is used.

    """

    _convert_null_values = field_validator("*")(utils.convert_null_values)



"""``runway gen-sample k8s-tf`` command."""

import logging
import shutil
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

import click

from ... import options
from .utils import TEMPLATES, convert_gitignore, copy_sample, write_tfstate_template

if TYPE_CHECKING:
    from ...._logging import RunwayLogger

LOGGER = cast("RunwayLogger", logging.getLogger(__name__.replace("._", ".")))


@click.command("k8s-tf-repo", short_help="k8s + tf (k8s-tf-infrastructure)")
@options.debug
@options.no_color
@options.verbose
@click.pass_context
def k8s_tf_repo(ctx: click.Context, **_: Any) -> None:
    """Generate a sample Terraform project using Kubernetes."""
    src = TEMPLATES / "k8s-tf-repo"
    dest = Path.cwd() / "k8s-tf-infrastructure"
    src_awscli = TEMPLATES / "k8s-cfn-repo/k8s-master.cfn/k8s_hooks/awscli.py"
    dest_awscli = dest / "gen-kubeconfig.cfn/k8s_hooks/awscli.py"

    copy_sample(ctx, src, dest)
    LOGGER.debug('copying "%s" to "%s"', src_awscli, dest_awscli)
    shutil.copyfile(src_awscli, dest_awscli)
    convert_gitignore(dest / "_gitignore")

    tfstate_dir = dest / "tfstate.cfn/templates"
    tfstate_dir.mkdir()
    write_tfstate_template(tfstate_dir / "tf_state.yml", bucket_deletion_policy="Delete")

    LOGGER.success("Sample k8s infrastructure repo created at %s", dest)
    LOGGER.notice("See the README for setup and deployment instructions.")



"""AWS KMS lookup."""

from __future__ import annotations

import codecs
import logging
from typing import TYPE_CHECKING, Any, BinaryIO, ClassVar, cast

from ....lookups.handlers.base import LookupHandler
from ....utils import DOC_SITE
from ...utils import read_value_from_path

if TYPE_CHECKING:
    from ....context import CfnginContext
    from ....lookups.handlers.base import ParsedArgsTypeDef

LOGGER = logging.getLogger(__name__)


class KmsLookup(LookupHandler["CfnginContext"]):
    """AWS KMS lookup."""

    DEPRECATION_MSG = (
        'lookup query syntax "<region>@<encrypted-blob>" has been deprecated; '
        "to learn how to use the new lookup query syntax visit "
        f"{DOC_SITE}/page/cfngin/lookups/kms.html"
    )
    TYPE_NAME: ClassVar[str] = "kms"
    """Name that the Lookup is registered as."""

    @classmethod
    def legacy_parse(cls, value: str) -> tuple[str, ParsedArgsTypeDef]:
        """Retain support for legacy lookup syntax.

        Format of value::

            <region>@<encrypted-blob>

        """
        LOGGER.warning("${%s %s}: %s", cls.TYPE_NAME, value, cls.DEPRECATION_MSG)
        region, value = read_value_from_path(value).split("@", 1)
        return value, {"region": region}

    @classmethod
    def handle(cls, value: str, context: CfnginContext, **_: Any) -> str:
        r"""Decrypt the specified value with a master key in KMS.

        Args:
            value: Parameter(s) given to this lookup.
            context: Context instance.

        """
        if "@" in value:
            query, args = cls.legacy_parse(value)
        else:
            query, args = cls.parse(value)

        kms = context.get_session(region=args.get("region")).client("kms")

        decrypted = cast(
            "BinaryIO | bytes",
            kms.decrypt(CiphertextBlob=codecs.decode(query.encode(), "base64")).get(
                "Plaintext", b""
            ),
        )
        if isinstance(decrypted, bytes):
            return cls.format_results(decrypted.decode(), **args)
        return cls.format_results(decrypted.read().decode(), **args)



"""Inverse of :class:`~pydantic.types.StrictStr`."""

from __future__ import annotations

from decimal import Decimal

from pydantic.functional_validators import BeforeValidator


def _handler(value: object | None) -> object | None:
    """Convert the provided value if able."""
    if isinstance(value, (float, int, Decimal)):
        return str(value)
    return value


LaxStr = BeforeValidator(_handler)
"""Custom :class:`~pydantic.functional_validators.BeforeValidator`.

Inverse of :class:`~pydantic.types.StrictStr` that allows additional types to be
accepted as a :class:`str`.

.. rubric:: Example
.. code-block:: python

    from __future__ import annotations

    from typing import Annotated

    from runway.utils import LaxStr

    class MyModel(BaseModel):
        some_field: Annotated[str, LaxStr]
        some_other_field: Annotated[str | None, LaxStr] = None

"""



"""``runway deploy`` command."""

# docs: file://./../../../docs/source/commands.rst
import logging
from typing import Any

import click
from pydantic import ValidationError

from ...core import Runway
from ...exceptions import ConfigNotFound, VariablesFileNotFound
from .. import options
from ..utils import select_deployments

LOGGER = logging.getLogger(__name__.replace("._", "."))


@click.command("deploy", short_help="deploy things")
@options.ci
@options.debug
@options.deploy_environment
@options.no_color
@options.tags
@options.verbose
@click.pass_context
def deploy(ctx: click.Context, debug: bool, tags: tuple[str, ...], **_: Any) -> None:
    """Deploy infrastructure as code.

    \b
    Process
    -------
    1. Determines the deploy environment.
        - "-e, --deploy-environment" option
        - "DEPLOY_ENVIRONMENT" environment variable
        - git branch name
            - strips "ENV-" prefix, master is converted to common
            - ignored if "ignore_git_branch: true"
        - name of the current working directory
    2. Selects deployments & modules to deploy.
        - (default) prompts
        - (tags) module contains all tags
        - (non-interactive) all
    3. Deploys selected deployments/modules in the order defined.

    """  # noqa: D301
    try:
        Runway(ctx.obj.runway_config, ctx.obj.get_runway_context()).deploy(
            select_deployments(ctx, ctx.obj.runway_config.deployments, tags)
        )
    except ValidationError as err:
        LOGGER.error(err, exc_info=debug)
        ctx.exit(1)
    except (ConfigNotFound, VariablesFileNotFound) as err:
        LOGGER.error(err.message, exc_info=debug)
        ctx.exit(1)



"""Handler for the ``path`` field of a Runway module."""

from __future__ import annotations

import logging
import re
from pathlib import Path
from typing import TYPE_CHECKING, ClassVar
from urllib.parse import parse_qs

from typing_extensions import TypedDict

from ...compat import cached_property
from ...config.components.runway import RunwayModuleDefinition
from ...config.models.runway import RunwayModuleDefinitionModel
from ...sources.git import Git
from ._deploy_environment import DeployEnvironment

if TYPE_CHECKING:
    from ...sources.source import Source

LOGGER = logging.getLogger(__name__)


class ModulePathMetadataTypeDef(TypedDict):
    """Type definition for ModulePath.metadata."""

    arguments: dict[str, str]
    cache_dir: Path
    location: str
    source: str
    uri: str


class ModulePath:
    """Handler for the ``path`` field of a Runway module."""

    ARGS_REGEX: ClassVar[str] = r"(\?)(?P<args>.*)$"
    REMOTE_SOURCE_HANDLERS: ClassVar[dict[str, type[Source]]] = {"git": Git}
    SOURCE_REGEX: ClassVar[str] = r"(?P<source>[a-z]+)(\:\:)"
    URI_REGEX: ClassVar[str] = r"(?P<uri>[a-z]+://[a-zA-Z0-9\./-]+?(?=//|\?|$))"

    def __init__(
        self,
        definition: Path | str | None = None,
        *,
        cache_dir: Path,
        deploy_environment: DeployEnvironment | None = None,
    ) -> None:
        """Instantiate class.

        Args:
            definition: Path definition.
            cache_dir: Directory to use for caching if needed.
            deploy_environment: Current deploy environment object.

        """
        self.cache_dir = cache_dir
        self.definition = definition or Path.cwd()
        self.env = deploy_environment or DeployEnvironment()

    @cached_property
    def arguments(self) -> dict[str, str]:
        """Remote source arguments."""
        if isinstance(self.definition, str):
            match = re.match(rf"^.*{self.ARGS_REGEX}", self.definition)
            if match:
                return {k: ",".join(v) for k, v in parse_qs(match.group("args")).items()}
        return {}

    @cached_property
    def location(self) -> str:
        """Location of the module."""
        if isinstance(self.definition, str):
            if re.match(r"^(/|//|\.|\./)", self.definition) or "::" not in self.definition:
                return re.sub(self.ARGS_REGEX, "", self.definition)
            no_src = re.sub(rf"^{self.SOURCE_REGEX}", "", self.definition)
            no_uri = re.sub(rf"^{self.URI_REGEX}", "", no_src)
            match = re.search(r"//(?P<location>[^\?\n]*)", no_uri)
            if match:
                return match.group("location")
        return "./"

    @cached_property
    def metadata(self) -> ModulePathMetadataTypeDef:
        """Information that describes the module path."""
        return {
            "arguments": self.arguments,
            "cache_dir": self.cache_dir,
            "location": self.location,
            "source": self.source,
            "uri": self.uri,
        }

    @cached_property
    def module_root(self) -> Path:
        """Root directory of the module."""
        if isinstance(self.definition, Path):
            return self.definition
        if self.source != "local":
            return self._fetch_remote_source()
        return self.env.root_dir / self.location

    @cached_property
    def source(self) -> str:
        """Source of the module."""
        if isinstance(self.definition, str):
            match = re.match(rf"^{self.SOURCE_REGEX}.*$", self.definition)
            if match:
                return match.group("source")
        return "local"

    @cached_property
    def uri(self) -> str:
        """Remote source URI."""
        if isinstance(self.definition, str):
            match = re.match(rf"^{self.SOURCE_REGEX}{self.URI_REGEX}", self.definition)
            if match:
                return match.group("uri")
        return ""

    def _fetch_remote_source(self) -> Path:
        """Fetch remote module source.

        Raises:
            NotImplementedError: The supplied source does not have a handler.

        """
        try:
            return self.REMOTE_SOURCE_HANDLERS[self.source](**self.metadata).fetch()
        except KeyError:
            raise NotImplementedError(
                f"{self.source} is not a supported Runway module source"
            ) from None

    @classmethod
    def parse_obj(
        cls,
        obj: Path | RunwayModuleDefinition | RunwayModuleDefinitionModel | str | None,
        *,
        cache_dir: Path,
        deploy_environment: DeployEnvironment | None = None,
    ) -> ModulePath:
        """Parse object.

        Args:
            obj: Object to parse.
            cache_dir: Directory to use for caching if needed.
            deploy_environment: Current deploy environment object.

        Raises:
            TypeError: Unsupported type provided.

        """
        if isinstance(obj, (RunwayModuleDefinition, RunwayModuleDefinitionModel)):
            return cls(
                cache_dir=cache_dir,
                definition=obj.path,
                deploy_environment=deploy_environment,
            )
        if isinstance(obj, (type(None), Path, str)):
            return cls(
                cache_dir=cache_dir,
                definition=obj,
                deploy_environment=deploy_environment,
            )
        raise TypeError(
            f"object type {type(obj)}; expected pathlib.Path, "
            "RunwayModuleDefinition, RunwayModuleDefinitionModel, or str"
        )



"""Runway Terraform Module options."""

from __future__ import annotations

from typing import Annotated

from pydantic import ConfigDict, Field, field_validator

from ...base import ConfigProperty


class RunwayTerraformArgsDataModel(ConfigProperty):
    """Model for Runway Terraform Module args option."""

    model_config = ConfigDict(
        extra="forbid",
        title="Runway Terraform Module args option",
        validate_default=True,
        validate_assignment=True,
    )

    apply: list[str] = []
    init: list[str] = []
    plan: list[str] = []


class RunwayTerraformBackendConfigDataModel(ConfigProperty):
    """Model for Runway Terraform Module terraform_backend_config option."""

    model_config = ConfigDict(
        extra="forbid",
        title="Runway Terraform Module terraform_backend_config option",
        validate_default=True,
        validate_assignment=True,
    )

    bucket: str | None = None
    dynamodb_table: str | None = None
    region: str | None = None
    workspace_key_prefix: str | None = None

    def __bool__(self) -> bool:
        """Evaluate the boolean value of the object instance."""
        data = self.model_dump(exclude_none=True)
        return "bucket" in data or "dynamodb_table" in data


class RunwayTerraformModuleOptionsDataModel(ConfigProperty):
    """Model for Runway Terraform Module options."""

    model_config = ConfigDict(
        extra="ignore",
        title="Runway Terraform Module options",
        populate_by_name=True,
        validate_default=True,
        validate_assignment=True,
    )

    args: RunwayTerraformArgsDataModel = RunwayTerraformArgsDataModel()
    backend_config: RunwayTerraformBackendConfigDataModel = Field(
        default=RunwayTerraformBackendConfigDataModel(),
        alias="terraform_backend_config",
    )
    version: Annotated[str | None, Field(alias="terraform_version")] = None
    workspace: Annotated[str | None, Field(alias="terraform_workspace")] = None
    write_auto_tfvars: Annotated[bool, Field(alias="terraform_write_auto_tfvars")] = False

    @field_validator("args", mode="before")
    @classmethod
    def _convert_args(cls, v: list[str] | dict[str, list[str]]) -> dict[str, list[str]]:
        """Convert args from list to dict."""
        if isinstance(v, list):
            return {"apply": v}
        return v



"""CLI utils."""

from __future__ import annotations

import logging
import os
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any

import click
import yaml

from ..compat import cached_property
from ..config import RunwayConfig
from ..context import RunwayContext
from ..core.components import DeployEnvironment
from ..exceptions import ConfigNotFound

if TYPE_CHECKING:
    from collections.abc import Iterator

    from ..config.components.runway import (
        RunwayDeploymentDefinition,
        RunwayModuleDefinition,
    )

LOGGER = logging.getLogger(__name__)


class CliContext:
    """CLI context object."""

    def __init__(
        self,
        *,
        ci: bool = False,
        debug: int = 0,
        deploy_environment: str | None = None,
        verbose: bool = False,
        **_: Any,
    ) -> None:
        """Instantiate class.

        Args:
            ci: Whether Runway is being run in non-interactive mode.
            debug: Debug level
            deploy_environment: Name of the deploy environment.
            verbose: Whether to display verbose logs.

        """
        self._deploy_environment = deploy_environment
        self.ci = ci
        self.debug = debug
        self.root_dir = Path.cwd()
        self.verbose = verbose

    @cached_property
    def env(self) -> DeployEnvironment:
        """Name of the current deploy environment."""
        environ = os.environ.copy()
        # carefully update environ with values passed from the cli
        if self.ci and "CI" not in environ:
            environ["CI"] = "1"
        if self.debug and "DEBUG" not in environ:
            environ["DEBUG"] = str(self.debug)
        if self.verbose and "VERBOSE" not in environ:
            environ["VERBOSE"] = "1"
        return DeployEnvironment(
            environ=environ,
            explicit_name=self._deploy_environment,
            root_dir=self.root_dir,
        )

    @cached_property
    def runway_config(self) -> RunwayConfig:
        """Runway config."""
        config = RunwayConfig.parse_file(file_path=self.runway_config_path)
        self.env.ignore_git_branch = config.ignore_git_branch
        return config

    @cached_property
    def runway_config_path(self) -> Path:
        """Path to the runway config file.

        Raises:
            SystemExit: Config file not found or multiple were matches were found.

        """
        try:
            path = RunwayConfig.find_config_file(self.root_dir)
            self.root_dir = path.parent
            self.env.root_dir = self.root_dir
            return path
        except ConfigNotFound as err:
            LOGGER.error(err.message)
        except ValueError as err:
            LOGGER.error(err)
        sys.exit(1)

    def get_runway_context(
        self, deploy_environment: DeployEnvironment | None = None
    ) -> RunwayContext:
        """Get a Runway context object.

        Uses the location of the config file to set the working directory for Runway.

        Args:
            deploy_environment: Object representing the current deploy environment.

        Returns:
            RunwayContext

        """
        return RunwayContext(
            deploy_environment=deploy_environment or self.env,
            work_dir=self.runway_config_path.parent / ".runway",
        )

    def __getitem__(self, key: str) -> Any:
        """Implement evaluation of self[key].

        Args:
            key: Attribute name to return the value for.

        Returns:
            The value associated with the provided key/attribute name.

        Raises:
            Attribute: If attribute does not exist on this object.

        """
        # ignore coverage for standard implementation
        return getattr(self, key)  # cov: ignore

    def __setitem__(self, key: str, value: Any) -> None:
        """Implement assignment to self[key].

        Args:
            key: Attribute name to associate with a value.
            value: Value of a key/attribute.

        """
        # ignore coverage for standard implementation
        setattr(self, key, value)  # cov: ignore

    def __delitem__(self, key: str) -> None:
        """Implement deletion of self[key].

        Args:
            key: Attribute name to remove from the object.

        """
        # ignore coverage for standard implementation
        delattr(self, key)  # cov: ignore

    def __len__(self) -> int:
        """Implement the built-in function len()."""
        # ignore coverage for standard implementation
        return len(self.__dict__)  # cov: ignore

    def __iter__(self) -> Iterator[str]:
        """Return iterator object that can iterate over all attributes."""
        # ignore coverage for standard implementation
        return iter(self.__dict__)  # cov: ignore

    def __str__(self) -> str:
        """Return string representation of the object."""
        # ignore coverage for standard implementation
        return f"CliContext({self.__dict__})"  # cov: ignore


def select_deployments(
    ctx: click.Context,
    deployments: list[RunwayDeploymentDefinition],
    tags: tuple[str, ...] | None = None,
) -> list[RunwayDeploymentDefinition]:
    """Select which deployments to run.

    Uses tags, interactive prompts, or selects all.

    Args:
        ctx: Current click context.
        deployments: List of deployment(s) to choose from.
        tags: Deployment tags to filter.

    Returns:
        Selected deployment(s).

    """
    if tags:
        return select_modules_using_tags(ctx, deployments, tags)
    if ctx.obj.env.ci:
        return deployments
    if len(deployments) == 1:
        choice = 1
        LOGGER.debug("only one deployment detected; no selection necessary")
    else:
        # build the menu before displaying it so debug logs don't break up what is printed
        deployment_menu = yaml.safe_dump({i + 1: d.menu_entry for i, d in enumerate(deployments)})
        click.secho("\nConfigured deployments\n", bold=True, underline=True)
        click.echo(deployment_menu)
        if ctx.command.name == "destroy":
            click.echo(
                '(operating in destroy mode -- "all" will destroy all '
                "deployments in reverse order)\n"
            )
        choice = click.prompt(
            'Enter number of deployment to run (or "all")',
            default="all",
            show_choices=False,
            type=click.Choice([str(n) for n in range(1, len(deployments) + 1)] + ["all"]),
        )
    if choice != "all":
        deployments = [deployments[int(choice) - 1]]
        deployments[0].modules = select_modules(ctx, deployments[0].modules)
    return deployments


def select_modules(
    ctx: click.Context, modules: list[RunwayModuleDefinition]
) -> list[RunwayModuleDefinition]:
    """Interactively select which modules to run.

    Args:
        ctx: Current click context.
        modules: List of module(s) to choose from.

    Returns:
        Selected module(s).

    """
    if len(modules) == 1:
        LOGGER.debug("only one module detected; no selection necessary")
        if ctx.command.name == "destroy":
            LOGGER.info(
                "Only one module detected; all modules automatically selected for deletion."
            )
            if not click.confirm("Proceed?"):
                ctx.exit(0)
        return modules
    click.secho("\nConfigured modules\n", bold=True, underline=True)
    click.echo(yaml.safe_dump({i + 1: m.menu_entry for i, m in enumerate(modules)}))
    if ctx.command.name == "destroy":
        click.echo(
            '(operating in destroy mode -- "all" will destroy all modules in reverse order)\n'
        )
    choice = click.prompt(
        'Enter number of module to run (or "all")',
        default="all",
        show_choices=False,
        type=click.Choice([str(n) for n in range(1, len(modules) + 1)] + ["all"]),
    )
    click.echo("")
    if choice == "all":
        return modules
    modules = [modules[int(choice) - 1]]
    if modules[0].child_modules:
        return select_modules(ctx, modules[0].child_modules)
    return modules


def select_modules_using_tags(
    ctx: click.Context,
    deployments: list[RunwayDeploymentDefinition],
    tags: tuple[str, ...],
) -> list[RunwayDeploymentDefinition]:
    """Select modules to run using tags.

    Args:
        ctx: Current click context.
        deployments: List of deployments to check.
        tags: List of tags to filter modules.

    Returns:
        List of selected deployments with selected modules.

    """
    deployments_to_run: list[RunwayDeploymentDefinition] = []
    for deployment in deployments:
        modules_to_run: list[RunwayModuleDefinition] = []
        for module in deployment.modules:
            if module.child_modules:
                module.child_modules = [
                    c for c in module.child_modules if all(t in c.tags for t in tags)
                ]
                if module.child_modules:
                    modules_to_run.append(module)
            elif all(t in module.tags for t in tags):
                modules_to_run.append(module)
        if modules_to_run:
            deployment.modules = modules_to_run
            deployments_to_run.append(deployment)
    if deployments_to_run:
        return deployments_to_run
    LOGGER.error("No modules found with the provided tag(s): %s", ", ".join(tags))
    return ctx.exit(1)



"""Runway config test definition."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar

from ....variables import Variable
from ...models.runway import RunwayTestDefinitionModel, ValidRunwayTestTypeValues
from .base import ConfigComponentDefinition

if TYPE_CHECKING:
    from typing_extensions import Self


class RunwayTestDefinition(ConfigComponentDefinition[RunwayTestDefinitionModel]):
    """Runway test definition."""

    args: dict[str, Any]
    name: str
    required: bool
    type: ClassVar[ValidRunwayTestTypeValues]

    _supports_vars: tuple[str, ...] = ("args", "required")

    def __init__(self, data: RunwayTestDefinitionModel) -> None:
        """Instantiate class."""
        super().__init__(data)

    def _register_variable(self, var_name: str, var_value: Any) -> None:
        """Register a variable.

        Args:
            var_name: Name of the config field that can contain a variable
                lookup.
            var_value: Literal value supplied in the config to be resolved
                as a variable if it contains a lookup.

        """
        self._vars[var_name] = Variable(
            name=f"{self.name}.{var_name}", value=var_value, variable_type="runway"
        )

    @classmethod
    def parse_obj(cls: type[Self], obj: object) -> Self:
        """Parse a python object into this class.

        Args:
            obj: The object to parse.

        """
        return cls(RunwayTestDefinitionModel.model_validate(obj))



"""CFNgin hook for building static website."""

from __future__ import annotations

import logging
import os
import tempfile
import zipfile
from pathlib import Path
from typing import TYPE_CHECKING, Any

import boto3
from boto3.s3.transfer import S3Transfer
from typing_extensions import TypedDict

from ....module.staticsite.options import RunwayStaticSiteSourceHashingDataModel
from ....s3_utils import does_s3_object_exist, download_and_extract_to_mkdtemp
from ....utils import change_dir, run_commands
from ...lookups.handlers.rxref import RxrefLookup
from ..base import HookArgsBaseModel
from .utils import get_hash_of_files

if TYPE_CHECKING:
    from ....cfngin.providers.aws.default import Provider
    from ....context import CfnginContext

LOGGER = logging.getLogger(__name__)


class HookArgsOptions(HookArgsBaseModel):
    """Hook arguments ``options`` block."""

    build_output: str | None = None
    """Path were the build static site will be stored locally before upload."""

    build_steps: list[str | list[str] | dict[str, str | list[str]]] = []
    """Steps to execute to build the static site."""

    name: str = "undefined"
    """Static site name."""

    namespace: str
    """Namespace of the static site."""

    path: str
    """Working directory/path to the static site's source code."""

    pre_build_steps: list[str | list[str] | dict[str, str | list[str]]] = []
    """Steps to run before building the static site."""

    source_hashing: RunwayStaticSiteSourceHashingDataModel = (
        RunwayStaticSiteSourceHashingDataModel()
    )
    """Settings for tracking the hash of the source code between runs."""


class HookArgs(HookArgsBaseModel):
    """Hook arguments."""

    artifact_bucket_rxref_lookup: str
    """Query for ``RxrefLookup`` to get artifact bucket."""

    options: HookArgsOptions
    """Hook ``options`` block."""


def zip_and_upload(
    app_dir: str, bucket: str, key: str, session: boto3.Session | None = None
) -> None:
    """Zip built static site and upload to S3."""
    s3_client = session.client("s3") if session else boto3.client("s3")
    transfer = S3Transfer(s3_client)

    filedes, temp_file = tempfile.mkstemp()
    os.close(filedes)
    LOGGER.info("archiving %s to s3://%s/%s", app_dir, bucket, key)
    with zipfile.ZipFile(temp_file, "w", zipfile.ZIP_DEFLATED) as filehandle, change_dir(app_dir):
        for dirname, _subdirs, files in os.walk("./"):
            if dirname != "./":
                filehandle.write(dirname)
            for filename in files:
                filehandle.write(os.path.join(dirname, filename))  # noqa: PTH118
    transfer.upload_file(temp_file, bucket, key)
    os.remove(temp_file)  # noqa: PTH107


class OptionsArgTypeDef(TypedDict, total=False):
    """Options argument type definition."""

    build_output: str
    build_steps: list[str | list[str] | dict[str, str | list[str]]]
    name: str
    namespace: str
    path: str
    pre_build_steps: list[str | list[str] | dict[str, str | list[str]]]


def build(
    context: CfnginContext,
    provider: Provider,
    *,
    options: OptionsArgTypeDef | None = None,
    **kwargs: Any,
) -> dict[str, Any]:
    """Build static site.

    Arguments parsed by :class:`~runway.cfngin.hooks.staticsite.build_staticsite.HookArgs`.

    """
    options = options or {}
    options.setdefault("namespace", context.namespace)
    options.setdefault("path", str(context.config_path))
    args = HookArgs.model_validate({"options": options, **kwargs})
    session = context.get_session()

    context_dict: dict[str, Any] = {
        "artifact_key_prefix": f"{args.options.namespace}-{args.options.name}-"
    }

    if args.options.build_output:
        build_output = os.path.join(args.options.path, args.options.build_output)  # noqa: PTH118
    else:
        build_output = args.options.path

    context_dict["artifact_bucket_name"] = RxrefLookup.handle(
        args.artifact_bucket_rxref_lookup, provider=provider, context=context
    )

    if args.options.pre_build_steps:
        run_commands(args.options.pre_build_steps, args.options.path)

    context_dict["hash"] = get_hash_of_files(
        root_path=Path(args.options.path),
        directories=options.get("source_hashing", {"directories": None}).get("directories"),
    )
    LOGGER.debug("application hash: %s", context_dict["hash"])

    # Now determine if the current staticsite has already been deployed
    if args.options.source_hashing.enabled:
        context_dict["hash_tracking_parameter"] = (
            args.options.source_hashing.parameter or f"{context_dict['artifact_key_prefix']}hash"
        )

        ssm_client = session.client("ssm")

        try:
            old_parameter_value = ssm_client.get_parameter(
                Name=context_dict["hash_tracking_parameter"]
            )["Parameter"].get("Value")
        except ssm_client.exceptions.ParameterNotFound:
            old_parameter_value = None
    else:
        context_dict["hash_tracking_disabled"] = True
        old_parameter_value = None

    context_dict["current_archive_filename"] = (
        context_dict["artifact_key_prefix"] + context_dict["hash"] + ".zip"
    )
    if old_parameter_value:
        context_dict["old_archive_filename"] = (
            context_dict["artifact_key_prefix"] + old_parameter_value + ".zip"
        )

    if old_parameter_value == context_dict["hash"]:
        LOGGER.info("skipped build; hash already deployed")
        context_dict["deploy_is_current"] = True
        return context_dict

    if does_s3_object_exist(
        context_dict["artifact_bucket_name"],
        context_dict["current_archive_filename"],
        session,
    ):
        context_dict["app_directory"] = download_and_extract_to_mkdtemp(
            context_dict["artifact_bucket_name"],
            context_dict["current_archive_filename"],
            session,
        )
    else:
        if args.options.build_steps:
            LOGGER.info("build steps (in progress)")
            run_commands(args.options.build_steps, args.options.path)
            LOGGER.info("build steps (complete)")
        zip_and_upload(
            build_output,
            context_dict["artifact_bucket_name"],
            context_dict["current_archive_filename"],
            session,
        )
        context_dict["app_directory"] = build_output

    context_dict["deploy_is_current"] = False
    return context_dict



"""Customized JSON encoder."""

from __future__ import annotations

import datetime
import json
from decimal import Decimal
from pathlib import Path
from typing import Any

from packaging.specifiers import SpecifierSet
from pydantic import BaseModel


class JsonEncoder(json.JSONEncoder):
    """Encode Python objects to JSON data.

    This class can be used with ``json.dumps()`` to handle most data types
    that can occur in responses from AWS.

    Usage:
        >>> json.dumps(data, cls=JsonEncoder)

    """

    def default(self, o: Any) -> dict[Any, Any] | float | list[Any] | str | Any:
        """Encode types not supported by the default JSONEncoder.

        Args:
            o: Object to encode.

        Returns:
            JSON serializable data type.

        Raises:
            TypeError: Object type could not be encoded.

        """
        if isinstance(o, Decimal):
            return float(o)
        if isinstance(o, (datetime.datetime, datetime.date)):
            return o.isoformat()
        if isinstance(o, BaseModel):
            return o.model_dump()
        if isinstance(o, (Path, SpecifierSet)):
            return str(o)
        if isinstance(o, (set, tuple)):
            return list(o)  # pyright: ignore[reportUnknownArgumentType]
        return super().default(o)



"""``runway dismantle`` command."""

# docs: file://./../../../docs/source/commands.rst
import logging
from typing import TYPE_CHECKING, Any, cast

import click

from .. import options
from ._destroy import destroy

if TYPE_CHECKING:
    from ..._logging import RunwayLogger

LOGGER = cast("RunwayLogger", logging.getLogger(__name__.replace("._", ".")))


@click.command("dismantle", short_help="alias of destroy")
@options.ci
@options.debug
@options.deploy_environment
@options.no_color
@options.tags
@options.verbose
@click.pass_context
def dismantle(ctx: click.Context, **kwargs: Any) -> None:
    """Alias of "runway destroy".

    For more information, refer to the output of "runway destroy --help".

    """
    LOGGER.verbose("forwarding to destroy...")
    ctx.forward(destroy, **kwargs)



"""Runway config components."""

from ._deployment_def import RunwayDeploymentDefinition
from ._module_def import RunwayModuleDefinition
from ._test_def import RunwayTestDefinition
from ._variables_def import RunwayVariablesDefinition

__all__ = [
    "RunwayDeploymentDefinition",
    "RunwayModuleDefinition",
    "RunwayTestDefinition",
    "RunwayVariablesDefinition",
]



"""Runway config module definition."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from ....variables import Variable
from ...models.runway import RunwayModuleDefinitionModel
from .base import ConfigComponentDefinition

if TYPE_CHECKING:
    from pathlib import Path

    from typing_extensions import Self

    from ...models.runway import RunwayEnvironmentsType, RunwayEnvVarsType, RunwayModuleTypeTypeDef


class RunwayModuleDefinition(ConfigComponentDefinition[RunwayModuleDefinitionModel]):
    """Runway module definition."""

    class_path: str | None
    environments: RunwayEnvironmentsType
    env_vars: RunwayEnvVarsType
    name: str
    options: dict[str, Any]
    parameters: dict[str, Any]
    path: str | Path | None
    tags: list[str]
    type: RunwayModuleTypeTypeDef | None

    _supports_vars: tuple[str, ...] = (
        "class_path",
        "env_vars",
        "environments",
        "options",
        "parameters",
        "path",
    )

    def __init__(self, data: RunwayModuleDefinitionModel) -> None:
        """Instantiate class."""
        super().__init__(data)

    @property
    def child_modules(self) -> list[RunwayModuleDefinition]:
        """List of child modules."""
        return [RunwayModuleDefinition(child) for child in self._data.parallel]

    @child_modules.setter
    def child_modules(
        self,
        modules: list[RunwayModuleDefinition | RunwayModuleDefinitionModel],
    ) -> None:
        """Set the value of the property.

        Args:
            modules: A list of modules.

        Raises:
            TypeError: The provided value does not match the required types.

        """
        if not isinstance(modules, list):  # type: ignore
            raise TypeError(f"expected list[RunwayModuleDefinition]; got {type(modules)}")
        sanitized: list[RunwayModuleDefinitionModel] = []
        for i, mod in enumerate(modules):
            if isinstance(mod, RunwayModuleDefinition):
                sanitized.append(RunwayModuleDefinitionModel.model_validate(mod.data))
            elif isinstance(mod, RunwayModuleDefinitionModel):  # type: ignore
                sanitized.append(mod)
            else:
                raise TypeError(
                    f"{self.name}.parallel[{i}] is type {type(mod)}; "
                    "expected type RunwayModuleDefinition or RunwayModuleDefinitionModel"
                )
        self._data.parallel = sanitized

    @property
    def is_parent(self) -> bool:
        """Assess if the modules contains child modules (e.g. run in parallel)."""
        return bool(self._data.parallel)

    @property
    def menu_entry(self) -> str:
        """Return menu entry representation of this module."""
        if self.is_parent:
            return f"{self.name} [{', '.join([c.menu_entry for c in self.child_modules])}]"
        return self.name

    def reverse(self) -> None:
        """Reverse the order of child/parallel modules."""
        self._data.parallel.reverse()

    def _register_variable(self, var_name: str, var_value: Any) -> None:
        """Register a variable.

        Args:
            var_name: Name of the config field that can contain a variable
                lookup.
            var_value: Literal value supplied in the config to be resolved
                as a variable if it contains a lookup.

        """
        self._vars[var_name] = Variable(
            name=f"{self.name}.{var_name}", value=var_value, variable_type="runway"
        )

    @classmethod
    def parse_obj(cls: type[Self], obj: object) -> Self:
        """Parse a python object into this class.

        Args:
            obj: The object to parse.

        """
        return cls(RunwayModuleDefinitionModel.model_validate(obj))



"""Retrieve a variable from the variables file or definition."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, ClassVar

from .base import LookupHandler

if TYPE_CHECKING:

    from ...utils import MutableMap


LOGGER = logging.getLogger(__name__)
TYPE_NAME = "var"


class VarLookup(LookupHandler[Any]):
    """Variable definition Lookup."""

    TYPE_NAME: ClassVar[str] = "var"
    """Name that the Lookup is registered as."""

    @classmethod
    def handle(cls, value: str, *_args: Any, variables: MutableMap, **_kwargs: Any) -> Any:
        """Retrieve a variable from the variable definition.

        The value is retrieved from the variables passed to Runway using
        either a variables file or the ``variables`` directive of the
        config file.

        Args:
            value: The value passed to the Lookup.
            variables: The resolved variables pass to Runway.

        Raises:
            ValueError: Unable to find a value for the provided query and
                a default value was not provided.

        """
        query, args = cls.parse(value)

        result = variables.find(query, default=args.pop("default", ""))

        if result != "":  # allows for False bool and NoneType results
            return cls.format_results(result, **args)

        raise ValueError(f'"{query}" does not exist in the variable definition')



"""CFNgin hook for cleaning up resources prior to CFN stack deletion."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from botocore.exceptions import ClientError

from ...utils import BaseModel

if TYPE_CHECKING:
    from ...context import CfnginContext

LOGGER = logging.getLogger(__name__)


class PurgeBucketHookArgs(BaseModel):
    """Hook arguments for ``purge_bucket``."""

    bucket_name: str
    """Name of the bucket to purge."""


def purge_bucket(context: CfnginContext, *__args: Any, **kwargs: Any) -> bool:
    """Delete objects in bucket."""
    args = PurgeBucketHookArgs.model_validate(kwargs)
    session = context.get_session()
    s3_resource = session.resource("s3")
    try:
        s3_resource.meta.client.head_bucket(Bucket=args.bucket_name)
    except ClientError as exc:
        if exc.response["Error"]["Code"] == "404":
            LOGGER.info('bucket "%s" does not exist; unable to complete purge', args.bucket_name)
            return True
        raise

    bucket = s3_resource.Bucket(args.bucket_name)
    bucket.object_versions.delete()
    return True



"""Split lookup."""

from typing import Any, ClassVar

from ....lookups.handlers.base import LookupHandler


class SplitLookup(LookupHandler[Any]):
    """Split lookup."""

    TYPE_NAME: ClassVar[str] = "split"
    """Name that the Lookup is registered as."""

    @classmethod
    def handle(cls, value: str, *_args: Any, **_kwargs: Any) -> list[str]:
        """Split the supplied string on the given delimiter, providing a list.

        Args:
            value: Parameter(s) given to this lookup.

        Format of value::

            <delimiter>::<value>

        Example:
            ::

                Subnets: ${split ,::subnet-1,subnet-2,subnet-3}

            Would result in the variable `Subnets` getting a list consisting
            of::

                ["subnet-1", "subnet-2", "subnet-3"]

            This is particularly useful when getting an output from another
            stack that contains a list. For example, the standard vpc blueprint
            outputs the list of Subnets it creates as a pair of Outputs
            (``PublicSubnets``, ``PrivateSubnets``) that are comma separated,
            so you could use this in your config::

                Subnets: ${split ,::${output vpc.PrivateSubnets}}

        """
        try:
            delimiter, text = value.split("::", 1)
        except ValueError:
            raise ValueError(
                f"Invalid value for split: {value}. Must be in <delimiter>::<text> format."
            ) from None

        return text.split(delimiter)



"""``runway gen-sample k8s-flux-repo`` command."""

import logging
import shutil
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

import click

from ... import options
from .utils import TEMPLATES, convert_gitignore, copy_sample, write_tfstate_template

if TYPE_CHECKING:
    from ...._logging import RunwayLogger

LOGGER = cast("RunwayLogger", logging.getLogger(__name__.replace("._", ".")))


@click.command("k8s-flux-repo", short_help="k8s + flux + tf (k8s-tf-infrastructure)")
@options.debug
@options.no_color
@options.verbose
@click.pass_context
def k8s_flux_repo(ctx: click.Context, **_: Any) -> None:
    """Generate a sample Kubernetes cluster with Flux CD managed via Terraform."""
    src = TEMPLATES / "k8s-flux-repo"
    dest = Path.cwd() / "k8s-tf-infrastructure"
    src_awscli = TEMPLATES / "k8s-cfn-repo/k8s-master.cfn/k8s_hooks/awscli.py"
    dest_awscli = dest / "gen-kubeconfig.cfn/k8s_hooks/awscli.py"

    copy_sample(ctx, src, dest)
    tf_eks_base = TEMPLATES / "k8s-tf-repo" / "eks-base.tf"
    copy_sample(ctx, tf_eks_base, dest / tf_eks_base.parts[-1])
    convert_gitignore(dest / "_gitignore")

    gen_kubeconfig_src_dir = TEMPLATES / "k8s-tf-repo" / "gen-kubeconfig.cfn"
    copy_sample(ctx, gen_kubeconfig_src_dir, dest / gen_kubeconfig_src_dir.parts[-1])
    LOGGER.debug('copying "%s" to "%s"', src_awscli, dest_awscli)
    shutil.copyfile(str(src_awscli), str(dest_awscli))

    tfstate_src_dir = TEMPLATES / "k8s-tf-repo" / "tfstate.cfn"
    copy_sample(ctx, tfstate_src_dir, dest / tfstate_src_dir.parts[-1])
    tfstate_templates_dir = dest / "tfstate.cfn/templates"
    tfstate_templates_dir.mkdir()
    write_tfstate_template(tfstate_templates_dir / "tf_state.yml", bucket_deletion_policy="Delete")

    LOGGER.success("Sample k8s infrastructure repo created at %s", dest)
    LOGGER.notice("See the README for setup and deployment instructions.")



"""Static website Module."""

from __future__ import annotations

import json
import logging
import os
import sys
import tempfile
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

import yaml

from ..._logging import PrefixAdaptor
from ...compat import cached_property
from ...utils import YamlDumper
from ..base import RunwayModule
from ..cloudformation import CloudFormation
from .options import StaticSiteOptions
from .parameters import RunwayStaticSiteModuleParametersDataModel
from .utils import add_url_scheme

if TYPE_CHECKING:
    from ..._logging import RunwayLogger
    from ...context import RunwayContext
    from ..base import ModuleOptions

LOGGER = cast("RunwayLogger", logging.getLogger(__name__))


class StaticSite(RunwayModule[StaticSiteOptions]):
    """Static website Runway Module."""

    DEPRECATION_MSG = (
        "Static website Runway module support has been deprecated and "
        "may be removed in the next major release."
    )

    parameters: RunwayStaticSiteModuleParametersDataModel

    def __init__(
        self,
        context: RunwayContext,
        *,
        explicitly_enabled: bool | None = False,
        logger: RunwayLogger = LOGGER,
        module_root: Path,
        name: str | None = None,
        options: dict[str, Any] | ModuleOptions | None = None,
        parameters: dict[str, Any] | None = None,
        **_: Any,
    ) -> None:
        """Instantiate class.

        Args:
            context: Runway context object for the current session.
            explicitly_enabled: Whether or not the module is explicitly enabled.
                This is can be set in the event that the current environment being
                deployed to matches the defined environments of the module/deployment.
            logger: Used to write logs.
            module_root: Root path of the module.
            name: Name of the module.
            options: Options passed to the module class from the config as ``options``
                or ``module_options`` if coming from the deployment level.
            parameters: Values to pass to the underlying infrastructure as code
                tool that will alter the resulting infrastructure being deployed.
                Used to templatize IaC.

        """
        super().__init__(
            context,
            explicitly_enabled=explicitly_enabled,
            logger=logger,
            module_root=module_root,
            name=name,
            options=StaticSiteOptions.parse_obj(options or {}),
            parameters=parameters,
        )
        self.parameters = RunwayStaticSiteModuleParametersDataModel.model_validate(self.parameters)
        # logger needs to be created here to use the correct logger
        self.logger = PrefixAdaptor(self.name, LOGGER)
        LOGGER.warning("%s:%s", self.name, self.DEPRECATION_MSG)
        self._ensure_valid_environment_config()
        self._ensure_cloudfront_with_auth_at_edge()
        self._ensure_correct_region_with_auth_at_edge()

    @cached_property
    def sanitized_name(self) -> str:
        """Sanitized name safe to use in a CloudFormation Stack name.

        Errors are usually caused here by a ``.`` in the name.
        This unintelligently replaces ``.`` with ``-``.

        If issues are still encountered, we can check against the regex of
        ``(?=^.{1,128}$)^[a-zA-Z][-a-zA-Z0-9_]+$``.

        """
        return self.name.replace(".", "-").strip("-")

    def deploy(self) -> None:
        """Create website CFN module and run CFNgin.deploy."""
        if self.parameters:
            if not self.parameters.cf_disable:
                self.logger.warning(
                    "initial creation of & updates to distributions can take "
                    "up to an hour to complete"
                )

                # Auth@Edge warning about subsequent deploys
                if (
                    self.parameters.auth_at_edge
                    and not self.parameters.aliases
                    and self.ctx.is_interactive
                ):
                    self.logger.warning(
                        "A hook that is part of the dependencies stack of "
                        "the Auth@Edge static site deployment is designed "
                        "to verify that the correct Callback URLs are "
                        "being used when a User Pool Client already "
                        "exists for the application. This ensures that "
                        "there is no interruption of service while the "
                        "deployment reaches the stage where the Callback "
                        "URLs are updated to that of the Distribution. "
                        "Because of this you may receive a change set "
                        "prompt on subsequent deploys."
                    )
            self._setup_website_module(command="deploy")
        else:
            self.logger.info("skipped; environment required but not defined")

    def destroy(self) -> None:
        """Create website CFN module and run CFNgin.destroy."""
        if self.parameters:
            self._setup_website_module(command="destroy")
        else:
            self.logger.info("skipped; environment required but not defined")

    def init(self) -> None:
        """Run init."""
        LOGGER.warning("init not currently supported for %s", self.__class__.__name__)

    def plan(self) -> None:
        """Create website CFN module and run CFNgin.diff."""
        if self.parameters:
            self._setup_website_module(command="plan")
        else:
            self.logger.info("skipped; environment required but not defined")

    def _setup_website_module(self, command: str) -> None:
        """Create CFNgin configuration for website module."""
        self.logger.info("generating CFNgin config...")
        module_dir = self._create_module_directory()
        self._create_dependencies_yaml(module_dir)
        self._create_staticsite_yaml(module_dir)

        # Earlier Runway versions included a CFN stack with a state machine
        # that attempted to automatically clean up the orphaned Lambda@Edge
        # functions. This was found to be unreliable and has been removed.
        # For a period of time (e.g. until the next major release) leaving this
        # in to automatically delete the stack. Not a major priority to have
        # Runway delete the old `-cleanup` stack, as the resources in it don't
        # have any costs when unused.
        if command == "destroy" and (
            self.parameters.auth_at_edge
            or self.parameters.model_dump().get("staticsite_rewrite_index_index")
        ):
            self._create_cleanup_yaml(module_dir)

        cfn = CloudFormation(
            self.ctx,
            explicitly_enabled=self.explicitly_enabled,
            module_root=module_dir,
            name=self.name,
            options=self.options.data.model_dump(),
            parameters=self.parameters.model_dump(by_alias=True),
        )
        self.logger.info("%s (in progress)", command)
        getattr(cfn, command)()
        self.logger.info("%s (complete)", command)

    def _create_module_directory(self) -> Path:
        module_dir = Path(tempfile.mkdtemp())
        self.logger.debug("using temporary directory: %s", module_dir)
        return module_dir

    def _create_dependencies_yaml(self, module_dir: Path) -> Path:
        """Create CFNgin config file for Static Site dependency stack.

        Resulting config file is save to ``module_dir`` as ``01-dependencies.yaml``.

        Args:
            module_dir: Path to the Runway module.

        Returns:
            Path to the file that was created.

        """
        pre_deploy: list[Any] = []

        pre_destroy = [
            {
                "args": {"bucket_name": f"${{rxref {self.sanitized_name}-dependencies::{i}}}"},
                "path": "runway.cfngin.hooks.cleanup_s3.purge_bucket",
                "required": True,
            }
            for i in ["AWSLogBucketName", "ArtifactsBucketName"]
        ]

        if self.parameters.auth_at_edge:
            if not self.parameters.aliases:
                # Retrieve the appropriate callback urls from the User Pool Client
                pre_deploy.append(
                    {
                        "args": {
                            "user_pool_arn": self.parameters.user_pool_arn,
                            "aliases": self.parameters.aliases,
                            "stack_name": f"${{namespace}}-{self.sanitized_name}-dependencies",
                        },
                        "data_key": "aae_callback_url_retriever",
                        "path": "runway.cfngin.hooks.staticsite.auth_at_edge."
                        "callback_url_retriever.get",
                        "required": True,
                    }
                )

            if self.parameters.create_user_pool:
                # Retrieve the user pool id
                pre_destroy.append(
                    {
                        "args": self._get_user_pool_id_retriever_variables(),
                        "data_key": "aae_user_pool_id_retriever",
                        "path": "runway.cfngin.hooks.staticsite.auth_at_edge."
                        "user_pool_id_retriever.get",
                        "required": True,
                    }
                )

                # Delete the domain prior to trying to delete the
                # User Pool Client that was created
                pre_destroy.append(
                    {
                        "args": self._get_domain_updater_variables(),
                        "data_key": "aae_domain_updater",
                        "path": "runway.cfngin.hooks.staticsite.auth_at_edge."
                        "domain_updater.delete",
                        "required": True,
                    }
                )
            else:
                # Retrieve the user pool id
                pre_deploy.append(
                    {
                        "args": self._get_user_pool_id_retriever_variables(),
                        "data_key": "aae_user_pool_id_retriever",
                        "path": "runway.cfngin.hooks.staticsite.auth_at_edge."
                        "user_pool_id_retriever.get",
                        "required": True,
                    }
                )

        content: dict[str, Any] = {
            "cfngin_bucket": "",
            "namespace": "${namespace}",
            "pre_deploy": pre_deploy,
            "pre_destroy": pre_destroy,
            "service_role": self.parameters.service_role,
            "stacks": {
                f"{self.sanitized_name}-dependencies": {
                    "class_path": "runway.blueprints.staticsite.dependencies.Dependencies",
                    "variables": self._get_dependencies_variables(),
                }
            },
        }

        out_file = module_dir / "01-dependencies.yaml"
        out_file.write_text(
            yaml.dump(content, default_flow_style=False, sort_keys=True), encoding="utf-8"
        )
        self.logger.debug("created %s:\n%s", out_file.name, yaml.dump(content, Dumper=YamlDumper))
        return out_file

    def _create_staticsite_yaml(self, module_dir: Path) -> Path:
        """Create CFNgin config file for Static Site.

        Resulting config file is save to ``module_dir`` as ``02-staticsite.yaml``.

        Args:
            module_dir: Path to the Runway module.

        Returns:
            Path to the file that was created.

        """
        # Default parameter name matches build_staticsite hook
        if not self.options.source_hashing.parameter:
            self.options.source_hashing.parameter = f"${{namespace}}-{self.sanitized_name}-hash"
        nonce_secret_param = f"${{namespace}}-{self.sanitized_name}-nonce-secret"

        build_staticsite_args: dict[str, Any] = {
            # ensures yaml.safe_load will work by using JSON to convert objects
            "options": json.loads(self.options.data.model_dump_json(by_alias=True))
        }
        build_staticsite_args["artifact_bucket_rxref_lookup"] = (
            f"{self.sanitized_name}-dependencies::ArtifactsBucketName"
        )
        build_staticsite_args["options"]["namespace"] = "${namespace}"
        build_staticsite_args["options"]["name"] = self.sanitized_name
        build_staticsite_args["options"]["path"] = str(self.ctx.env.root_dir.resolve() / self.path)

        site_stack_variables = self._get_site_stack_variables()

        class_path = "staticsite.StaticSite"

        pre_deploy = [
            {
                "args": build_staticsite_args,
                "data_key": "staticsite",
                "path": "runway.cfngin.hooks.staticsite.build_staticsite.build",
                "required": True,
            }
        ]

        post_deploy = [
            {
                "args": {
                    "bucket_name": f"${{cfn ${{namespace}}-{self.sanitized_name}.BucketName}}",
                    "cf_disabled": site_stack_variables["DisableCloudFront"],
                    "distribution_domain": f"${{cfn ${{namespace}}-{self.sanitized_name}."
                    "CFDistributionDomainName::default=undefined}",
                    "distribution_id": f"${{cfn ${{namespace}}-{self.sanitized_name}"
                    ".CFDistributionId::default=undefined}",
                    "extra_files": [i.model_dump() for i in self.options.extra_files],
                    "website_url": f"${{cfn ${{namespace}}-{self.sanitized_name}"
                    ".BucketWebsiteURL::default=undefined}",
                },
                "path": "runway.cfngin.hooks.staticsite.upload_staticsite.sync",
                "required": True,
            }
        ]

        pre_destroy = [
            {
                "args": {"bucket_name": f"${{rxref {self.sanitized_name}::BucketName}}"},
                "path": "runway.cfngin.hooks.cleanup_s3.purge_bucket",
                "required": True,
            }
        ]

        if self.parameters.rewrite_directory_index or self.parameters.auth_at_edge:
            pre_destroy.append(
                {
                    "args": {"stack_relative_name": self.sanitized_name},
                    "path": "runway.cfngin.hooks.staticsite.cleanup.warn",
                    "required": False,
                }
            )

        post_destroy = [
            {
                "args": {"parameter_name": i},
                "path": "runway.cfngin.hooks.cleanup_ssm.delete_param",
            }
            for i in [
                self.options.source_hashing.parameter,
                nonce_secret_param,
                f"{self.options.source_hashing.parameter}extra",
            ]
        ]

        if self.parameters.auth_at_edge:
            class_path = "auth_at_edge.AuthAtEdge"

            pre_deploy.append(
                {
                    "path": "runway.cfngin.hooks.staticsite.auth_at_edge."
                    "user_pool_id_retriever.get",
                    "required": True,
                    "data_key": "aae_user_pool_id_retriever",
                    "args": self._get_user_pool_id_retriever_variables(),
                }
            )
            pre_deploy.append(
                {
                    "path": "runway.cfngin.hooks.staticsite.auth_at_edge.domain_updater.update",
                    "required": True,
                    "data_key": "aae_domain_updater",
                    "args": self._get_domain_updater_variables(),
                }
            )
            pre_deploy.append(
                {
                    "path": "runway.cfngin.hooks.staticsite.auth_at_edge.lambda_config.write",
                    "required": True,
                    "data_key": "aae_lambda_config",
                    "args": self._get_lambda_config_variables(
                        site_stack_variables,
                        nonce_secret_param,
                        self.parameters.required_group,
                    ),
                }
            )
            if not self.parameters.aliases:
                post_deploy.insert(
                    0,
                    {
                        "path": "runway.cfngin.hooks.staticsite.auth_at_edge."
                        "client_updater.update",
                        "required": True,
                        "data_key": "client_updater",
                        "args": self._get_client_updater_variables(
                            self.sanitized_name, site_stack_variables
                        ),
                    },
                )

        if self.parameters.role_boundary_arn:
            site_stack_variables["RoleBoundaryArn"] = self.parameters.role_boundary_arn

        site_stack_variables["custom_error_responses"] = [
            i.model_dump(exclude_none=True) for i in self.parameters.custom_error_responses
        ]
        site_stack_variables["lambda_function_associations"] = [
            i.model_dump() for i in self.parameters.lambda_function_associations
        ]

        content: dict[str, Any] = {
            "cfngin_bucket": "",
            "namespace": "${namespace}",
            "post_deploy": post_deploy,
            "post_destroy": post_destroy,
            "pre_deploy": pre_deploy,
            "pre_destroy": pre_destroy,
            "service_role": self.parameters.service_role,
            "stacks": {
                self.sanitized_name: {
                    "class_path": f"runway.blueprints.staticsite.{class_path}",
                    "variables": site_stack_variables,
                }
            },
        }

        out_file = module_dir / "02-staticsite.yaml"
        out_file.write_text(
            yaml.dump(content, default_flow_style=False, sort_keys=True), encoding="utf-8"
        )
        self.logger.debug("created 02-staticsite.yaml:\n%s", yaml.dump(content, Dumper=YamlDumper))
        return out_file

    def _create_cleanup_yaml(self, module_dir: Path) -> Path:
        """Create CFNgin config file for Static Site cleanup stack.

        Resulting config file is save to ``module_dir`` as ``03-cleanup.yaml``.

        Args:
            module_dir: Path to the Runway module.

        Returns:
            Path to the file that was created.

        """
        content = {
            "namespace": "${namespace}",
            "cfngin_bucket": "",
            "service_role": self.parameters.service_role,
            "stacks": {
                f"{self.sanitized_name}-cleanup": {
                    "template_path": os.path.join(  # noqa: PTH118
                        tempfile.gettempdir(),
                        "thisfileisnotused.yaml",  # cspell: disable-line
                    ),
                }
            },
        }

        out_file = module_dir / "03-cleanup.yaml"
        out_file.write_text(
            yaml.dump(content, default_flow_style=False, sort_keys=True), encoding="utf-8"
        )
        self.logger.debug("created %s:\n%s", out_file.name, yaml.dump(content, Dumper=YamlDumper))
        return out_file

    def _get_site_stack_variables(self) -> dict[str, Any]:
        site_stack_variables: dict[str, Any] = {
            "Aliases": [],
            "Compress": self.parameters.compress,
            "DisableCloudFront": self.parameters.cf_disable,
            "RedirectPathAuthRefresh": "${default staticsite_redirect_path_auth_refresh::"
            "/refreshauth}",
            "RedirectPathSignIn": "${default staticsite_redirect_path_sign_in::/parseauth}",
            "RedirectPathSignOut": "${default staticsite_redirect_path_sign_out::/}",
            "RewriteDirectoryIndex": self.parameters.rewrite_directory_index or "",
            "SignOutUrl": "${default staticsite_sign_out_url::/signout}",
            "WAFWebACL": self.parameters.web_acl or "",
        }

        if self.parameters.aliases:
            site_stack_variables["Aliases"] = self.parameters.aliases

        if self.parameters.acmcert_arn:
            site_stack_variables["AcmCertificateArn"] = self.parameters.acmcert_arn

        if self.parameters.enable_cf_logging:
            site_stack_variables["LogBucketName"] = (
                f"${{rxref {self.sanitized_name}-dependencies::AWSLogBucketName}}"
            )

        if self.parameters.auth_at_edge:
            self._ensure_auth_at_edge_requirements()
            site_stack_variables["UserPoolArn"] = self.parameters.user_pool_arn
            site_stack_variables["NonSPAMode"] = self.parameters.non_spa
            site_stack_variables["HttpHeaders"] = self.parameters.http_headers
            site_stack_variables["CookieSettings"] = self.parameters.cookie_settings
            site_stack_variables["OAuthScopes"] = self.parameters.oauth_scopes
        else:
            site_stack_variables["custom_error_responses"] = [
                i.model_dump(exclude_none=True) for i in self.parameters.custom_error_responses
            ]
            site_stack_variables["lambda_function_associations"] = [
                i.model_dump() for i in self.parameters.lambda_function_associations
            ]

        return site_stack_variables

    def _get_dependencies_variables(self) -> dict[str, Any]:
        variables: dict[str, Any] = {"OAuthScopes": self.parameters.oauth_scopes}
        if self.parameters.auth_at_edge:
            self._ensure_auth_at_edge_requirements()

            variables.update(
                {
                    "AuthAtEdge": self.parameters.auth_at_edge,
                    "SupportedIdentityProviders": self.parameters.supported_identity_providers,
                    "RedirectPathSignIn": (
                        "${default staticsite_redirect_path_sign_in::/parseauth}"
                    ),
                    "RedirectPathSignOut": ("${default staticsite_redirect_path_sign_out::/}"),
                },
            )

            if self.parameters.aliases:
                variables.update({"Aliases": self.parameters.aliases})
            if self.parameters.additional_redirect_domains:
                variables.update(
                    {"AdditionalRedirectDomains": self.parameters.additional_redirect_domains}
                )
            if self.parameters.create_user_pool:
                variables.update({"CreateUserPool": self.parameters.create_user_pool})

        return variables

    def _get_user_pool_id_retriever_variables(self) -> dict[str, Any]:
        args: dict[str, Any] = {
            "user_pool_arn": self.parameters.user_pool_arn,
        }

        if self.parameters.create_user_pool:
            args["created_user_pool_id"] = (
                f"${{rxref {self.sanitized_name}-dependencies::AuthAtEdgeUserPoolId}}"
            )

        return args

    def _get_domain_updater_variables(self) -> dict[str, str]:
        return {
            "client_id_output_lookup": f"{self.sanitized_name}-dependencies::AuthAtEdgeClient",
            "client_id": f"${{rxref {self.sanitized_name}-dependencies::AuthAtEdgeClient}}",
        }

    def _get_lambda_config_variables(
        self,
        site_stack_variables: dict[str, Any],
        nonce_secret_param: str,
        required_group: str | None = None,
    ) -> dict[str, Any]:
        return {
            "client_id": f"${{rxref {self.sanitized_name}-dependencies::AuthAtEdgeClient}}",
            "bucket": f"${{rxref {self.sanitized_name}-dependencies::ArtifactsBucketName}}",
            "cookie_settings": site_stack_variables["CookieSettings"],
            "http_headers": site_stack_variables["HttpHeaders"],
            "nonce_signing_secret_param_name": nonce_secret_param,
            "oauth_scopes": site_stack_variables["OAuthScopes"],
            "redirect_path_refresh": site_stack_variables["RedirectPathAuthRefresh"],
            "redirect_path_sign_in": site_stack_variables["RedirectPathSignIn"],
            "redirect_path_sign_out": site_stack_variables["RedirectPathSignOut"],
            "required_group": required_group,
        }

    def _get_client_updater_variables(
        self, name: str, site_stack_variables: dict[str, Any]
    ) -> dict[str, Any]:
        return {
            "alternate_domains": [add_url_scheme(x) for x in site_stack_variables["Aliases"]],
            "client_id": f"${{rxref {self.sanitized_name}-dependencies::AuthAtEdgeClient}}",
            "distribution_domain": f"${{rxref {name}::CFDistributionDomainName}}",
            "oauth_scopes": site_stack_variables["OAuthScopes"],
            "redirect_path_sign_in": site_stack_variables["RedirectPathSignIn"],
            "redirect_path_sign_out": site_stack_variables["RedirectPathSignOut"],
            "supported_identity_providers": self.parameters.supported_identity_providers,
        }

    def _ensure_auth_at_edge_requirements(self) -> None:
        if not (self.parameters.user_pool_arn or self.parameters.create_user_pool):
            self.logger.error(
                "staticsite_user_pool_arn or staticsite_create_user_pool "
                "is required for Auth@Edge; "
            )
            sys.exit(1)

    def _ensure_correct_region_with_auth_at_edge(self) -> None:
        """Exit if not in the us-east-1 region and deploying to Auth@Edge.

        Lambda@Edge is only available within the us-east-1 region.

        """
        if self.parameters.auth_at_edge and self.region != "us-east-1":
            self.logger.error("Auth@Edge must be deployed in us-east-1.")
            sys.exit(1)

    def _ensure_cloudfront_with_auth_at_edge(self) -> None:
        """Exit if both the Auth@Edge and CloudFront disablement are true."""
        if self.parameters.cf_disable and self.parameters.auth_at_edge:
            self.logger.error(
                'staticsite_cf_disable must be "false" if staticsite_auth_at_edge is "true"'
            )
            sys.exit(1)

    def _ensure_valid_environment_config(self) -> None:
        """Exit if config is invalid."""
        if not self.parameters.namespace:
            self.logger.error("namespace parameter is required but not defined")
            sys.exit(1)
